/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 *
 * This video driver is STRONGLY derived from the video drivers of the
 * VSyncMAME project by S.Sakamaki (Saka)
 */

#include "compil.h"
#include "blade3d.h"
#include "card.h"

/* 3C4 */
#define OldNewModeControl 0x0B

#define NewMode2 0x0D
#define NewMode1 0x0E
#define Protection 0x11
#define MCLKLow 0x16
#define MCLKHigh 0x17
#define ClockLow 0x18
#define ClockHigh 0x19

/* 3x4 */
#define InterlacedVsyncAdjust 0x19
#define HorizontalParamOverflow 0x2B

#define Offset 0x13
#define CRTCModuleTest 0x1E
#define CRTHiOrd 0x27
#define AddColReg 0x29

/* 3CE */
#define MiscInternalControlReg 0x2F

struct blade3d_id {
	unsigned value;
	const char* name;
	int supported;
	long ref; /* kHz */
	long vco_min; /* kHz */
	long vco_max; /* kHz */
};

/* Configuration */
#define BLADE3D_CONF_UNSUPPORTED 0, 0, 0, 0
#define BLADE3D_CONF 1, 14318, 50000, 230000

static struct blade3d_id blade3d_id_list[] = {
{ 0x9880, "Trident Blade3D", BLADE3D_CONF },
{ 0, 0, BLADE3D_CONF_UNSUPPORTED }
};

static unsigned blade3d_bus_device_func;

static struct blade3d_id* blade3d_card; /* Blade3D detected */

static DWORD VCLK_MAX = 230000000;

/*
int detectBlade(void)
{
	__dpmi_regs 	r;

	r.d.eax = 0x0000b101;						// PCI BIOS - INSTALLATION CHECK
	r.d.edi = 0x00000000;
	__dpmi_int(0x1a, &r);
	if( r.d.edx != 0x20494350 ){				// ' ICP'
		logerror("15.75KHz: PCI BIOS not Installed.\n");
		return 0;
	}

	while(1){
		r.d.eax = 0x0000b102;						// PCI BIOS - FIND PCI DEVICE
		r.d.ecx = 0x00009880;						// device ID
		r.d.edx = 0x00001023;						// vendor ID
		r.d.esi = 0x00000000;						// device index
		__dpmi_int(0x1a, &r);
		if( r.h.ah == 0 ){
			if(DetectBlade3D() != 0){
				REFFREQ  =	14318180;
				VCO_MAX  = 230000000;
				VCO_MIN  = (DWORD)(REFFREQ * 3.5);		// It seems that Blade3D still has this limitation
				VCLK_MAX = 230000000;
				logerror("Trident Blade 3D found\n");
				break;
			}
		}
		logerror("no Trident Blade 3D found\n");
		return 0;
	}

#ifdef MAME
	DisableStretch();
#endif

	return 1;
}
*/

static void blade3d_doublescan_set(int flag)
{
	if(flag == 0){
		card_crtc_set(0x09, card_bitmov(card_crtc_get(0x09), 7, 0, 0));
	}else{
		card_crtc_set(0x09, card_bitmov(card_crtc_get(0x09), 7, 1, 0));
	}
}

static void blade3d_double_logical_line_width_set(int flag)
{
	DWORD d0;

	if (flag != 0) flag = 1;
	d0 = card_graph_get(MiscInternalControlReg);
	d0 = card_bitmov(d0, 2, flag, 0);
	card_graph_set(MiscInternalControlReg, (BYTE)d0);
}

static void blade3d_interlaced_set(int flag)
{
	BYTE status;

	status = card_crtc_get(CRTCModuleTest) & 0x04;
	if (flag == 0) {
		if (status != 0) {
			blade3d_double_logical_line_width_set(0);
			card_crtc_set(CRTCModuleTest, card_crtc_get(CRTCModuleTest) & ~0x04);
		}
	} else {
		if (status == 0) {
			blade3d_double_logical_line_width_set(1);
			card_crtc_set(CRTCModuleTest, card_crtc_get(CRTCModuleTest) | 0x04);
		}
	}
}

static void blade3d_ext_set(const card_crtc STACK_PTR* cp) {
	BYTE d0;
	int d1;
	int interlaceoffset;

	d1 = (cp->HSStart/8) + (cp->HSEnd/8) - ((cp->HTotal/8) - 5);
	if(d1 < 0) d1 = 0;
	d1 = (d1 * cp->interlaceratio) / 100;
	d1 = d1 / 2;
	interlaceoffset = d1;
	card_crtc_set(InterlacedVsyncAdjust, (BYTE)(interlaceoffset & 0xFF));

	d0 = card_crtc_get(HorizontalParamOverflow);
	d0 = card_bitmov(d0, 0, (cp->HTotal	>> 3) - 5, 9);
	d0 = card_bitmov(d0, 1, (cp->HDisp	>> 3) - 1, 9);
	d0 = card_bitmov(d0, 2, interlaceoffset		 , 9);
	d0 = card_bitmov(d0, 3, (cp->HSStart >> 3)	 , 9);
	d0 = card_bitmov(d0, 4, (cp->HBStart >> 3)	 , 9);
	card_crtc_set(HorizontalParamOverflow, d0);

	d0 = card_crtc_get(CRTHiOrd);
	d0 = card_bitmov(d0, 3,	1, 0); /* line compare bit 10 */
	d0 = card_bitmov(d0, 4, cp->VDisp - 1, 10);
	d0 = card_bitmov(d0, 5, cp->VSStart, 10);
	d0 = card_bitmov(d0, 6, cp->VBStart, 10);
	d0 = card_bitmov(d0, 7, cp->VTotal - 2, 10);
	card_crtc_set(CRTHiOrd, d0);

	blade3d_doublescan_set(cp->doublescan);
	card_char_size_y_set(1);
	blade3d_interlaced_set(cp->interlace);
	card_polarity_set(cp->hpolarity, cp->vpolarity);
}

/*
static DWORD GetBladeBytesPerLine(void)
{
	DWORD d0;

	d0 = card_crtc_get(Offset);
	d0 = card_bitmov(d0, 8, card_crtc_get(AddColReg), 4);
	d0 = card_bitmov(d0, 9, card_crtc_get(AddColReg), 5);
	d0 = d0 << 3;

	return d0;
}
*/

/*
static void SetBladeBytesPerLine(DWORD d0)
{
	BYTE d1;

	d0 = d0 >> 3;
	card_crtc_set(Offset, (BYTE)(d0 & 0xFF));
	d1 = card_crtc_get(AddColReg);
	d1 = (BYTE)card_bitmov(d1, 4, d0, 8);
	d1 = (BYTE)card_bitmov(d1, 5, d0, 9);
	card_crtc_set(AddColReg, d1);
}
*/

static long blade3d_clock_compute(long vclk, int STACK_PTR* Num, int STACK_PTR* DeN, int STACK_PTR* PS, int STACK_PTR* PCDiv) {
	double desfreq;
	double temp;
	double tempdt;
	double dt;
	int n, startn, endn;
	int m, startm, endm;
	int k, startk, endk;
	int pdindex, startpdindex, endpdindex;
	double pd;
	int powerup[4] = { 1, 2, 4, 8 };
	double pdd[4] = {1, 1.5, 2, 4 };
	double pdi[4] = {10, 15, 20, 40 };
	int d0;
	int invalid_n_value[] = {
#if 1
		 0,  1,  2,  3,  4,  5,  6,  7,
		 9, 10, 11, 12, 13, 14, 15,
		18, 19, 20, 21, 22, 23,
		27, 28, 29, 30, 31,
		36, 37, 38, 39,
		45, 46, 47,
		54, 55,
		63,
#endif
		-1
	};

	if(vclk > VCLK_MAX) vclk = VCLK_MAX;

	startn = 8;
	endn   = 247;
	startm = 2;
	endm   = 61;
	startk = 0;
	endk   = 3;
	startpdindex = 0;
	endpdindex	 = 3;

	desfreq = (double)vclk;
	dt = VCLK_MAX;
	*Num = 0;
	*DeN = 0;
	*PS  = 0;
	*PCDiv	= 0;

	for(pdindex = startpdindex; pdindex <= endpdindex; pdindex += 1){
		pd = pdd[pdindex];
		for(k = startk; k <= endk; k += 1){
			for(n = startn; n <= endn; n += 1){
				d0 = 0;
				while(1){
					if(invalid_n_value[d0] == -1) break;
					if(invalid_n_value[d0] == n) break;
					d0 += 1;
				}
				if(invalid_n_value[d0] != -1) continue;
				for(m = startm; m <= endm; m += 1){
					temp = blade3d_card->ref * 1000.0 * ((double)(n + 8) / (double)(m + 2));
					if(blade3d_card->vco_min * 1000.0 <= temp && temp <= blade3d_card->vco_max * 1000.0) {
						temp = blade3d_card->ref * 1000.0  * ((double)(n + 8) / (double)((m + 2) * powerup[k]));
						temp = temp / pd;
						tempdt = desfreq - temp;
						if(tempdt < 0) tempdt *= -1;
						if(tempdt < dt){
							*Num = n;
							*DeN = m;
							*PS  = k;
							*PCDiv	= pdi[pdindex];
							dt = tempdt;
						}
					}
				}
			}
		}
	}

	n = (*Num + 8);
	m = (*DeN + 2);
	k = powerup[*PS];
	pd = (double)(*PCDiv) / 10;

	CARD_LOG(("blade3d: Num %3d, DeN %3d, PS %1d, PCDiv %.1f\n", n, m, k, pd));
	CARD_LOG(("blade3d: %.5f * (Num/DeN) : %f\n", (double)blade3d_card->ref * 1000.0/1000000, (double)blade3d_card->ref * 1000.0/1000000 *(double)n / m));
	CARD_LOG(("blade3d: %.5f * (Num/DeN)/(PS * PCDiv): %f\n", (double)blade3d_card->ref * 1000.0/1000000, (double)blade3d_card->ref * 1000.0/1000000 * (double)n/(m * k * pd)));
	vclk = blade3d_card->ref * 1000.0 * (double)n/(m * k * pd);
	CARD_LOG(("blade3d: vclk : %f\n", (double)vclk / 1000000.0));

	return vclk;
}

static void blade3d_unprotect_registers(void)
{
	card_seq_set(Protection, 0x87);
	card_seq_set(NewMode1, 0xC0 ^ 0x02);
	card_crtc_set(CRTCModuleTest, card_crtc_get(CRTCModuleTest) & 0xBF);
}

static void blade3d_clock_divider_set(int d0) {
//
//	d0 = PixelClockDiv * 10
//

	BYTE d1;
	BYTE d2;

	switch(d0){
		case 10:		 // div 1
			d1 = 0;
			break;
		case 15:		 // div 1.5
			d1 = 3;
			break;
		case 20:		 // div 2
			d1 = 1;
			break;
		case 40:		 // div 4
			d1 = 2;
			break;
		default:
			d1 = 0;
	}

	d1 = d1 << 1;
	d1 = d1 &  0x06;
	d2 = card_seq_get(NewMode2);
	d2 = d2 & ~0x06;
	d2 = d2 | d1;
	card_seq_set(NewMode2, d2);
}

static void blade3d_clock_set(int Num, int DeN, int PS, int PCDiv) {
	BYTE d0;
	BYTE d1;
	BYTE d2;

	blade3d_clock_divider_set(PCDiv);
	card_divider_set(1);

	/* N is all 8bits */
	d1 = Num;
	/* M is first 6bits, with K last 2bits */
	d2 = (DeN & 0x3F) | (PS << 6);

	card_seq_set(ClockLow, d1);
	card_seq_set(ClockHigh, d2);

	d0 = (card_in(0x3CC) & 0xF3) | 0x08;
	card_out(0x3C2, d0);							// use vclk2
}

static void blade3d_new_mode(void) {
	BYTE d0;
	d0 = card_seq_get(OldNewModeControl);
	(void)d0;
	CARD_LOG(("blade3d: chip version %2Xh\n", d0));
}

/*
static double GetCurrentDotClock(void)
{
	double d0;
	BYTE d1;
	BYTE d2;
	int k;
	int m;
	int n;
	int powerup[4] = { 1, 2, 4, 8 };


	d1 = card_seq_get(ClockLow);
	d2 = card_seq_get(ClockHigh);

	k = (d2 >> 6) & 0x03;
	m = d2 & 0x3F;
	m += 2;
	n = d1;
	n += 8;

	d0 = REFFREQ * ((double)n / (m * powerup[k]));
	d0 = d0 / (GetPixelClockDivider() * GetSeqDivider());
	return d0;
}
*/

/*
static DWORD GetHTotal(void)
{
	DWORD d0;

	d0 = card_crtc_get(0x00);
	d0 = card_bitmov(d0, 9, card_crtc_get(HorizontalParamOverflow), 0);
	d0 += 5;
	d0 *= 8;

	return d0;
}
*/

/*
static DWORD GetVTotal(void)
{
	DWORD d0;
	BYTE d1;

	// Unlock
	card_crtc_set(0x11, card_bitmov(card_crtc_get(0x11), 7, 0, 0));

	d1 = card_crtc_get(0x03);
	d1 = card_bitmov(d1, 7, 1, 0);			// compatible read
	card_crtc_set(0x03, d1);

	d0 = card_crtc_get(0x06);
	d0 = card_bitmov(d0, 8, card_crtc_get(0x07), 0);
	d0 = card_bitmov(d0, 9, card_crtc_get(0x07), 5);

	d0 = card_bitmov(d0, 10, card_crtc_get(CRTHiOrd), 7);

	d0 += 2;

	return d0;
}
*/


/*
static double GetCurrentMClock(void)
{
	double d0;
	BYTE d1;
	BYTE d2;
	int k;
	int m;
	int n;
	int powerup[4] = { 1, 2, 4, 8 };


	d1 = card_seq_get(MCLKLow);
	d2 = card_seq_get(MCLKHigh);

	k = (d2 >> 6) & 0x03;
	m = d2 & 0x3F;
	m += 2;
	n = d1;
	n += 8;

	d0 = (double)(REFFREQ * n) / (m * powerup[k]);

	return d0/1000000;
}
*/

/*
static double GetPixelClockDivider(void)
{
	BYTE d0;
	double pd[4] = { 1, 2, 4, 1.5 };


	d0 = card_seq_get(NewMode2);
	d0 = d0 >> 1;
	d0 = d0 & 0x03;

	return pd[ (int)d0 ];
}
*/

/*
static int DetectBlade3D(void)
{
	BYTE d0;
	int result;

	d0 = card_seq_get(0x09);
//	  logerror("Seq index 09h %2X\n", d0);
	d0 = d0 & 0xF8;
	if(d0 == 0x38){
		result = 1;
	}else{
		result = 0;
	}

	return result;
}
*/

/*
DWORD GetBladeScreenStartAddress(void)
{
	DWORD d0;

	d0	= (DWORD)card_crtc_get(0x0D);
	d0 |= (DWORD)card_crtc_get(0x0C) << 8;
	d0	= card_bitmov(d0, 16, card_crtc_get(CRTCModuleTest), 5);
	d0	= card_bitmov(d0, 17, card_crtc_get(CRTHiOrd), 0);
	d0	= card_bitmov(d0, 18, card_crtc_get(CRTHiOrd), 1);
	d0	= card_bitmov(d0, 19, card_crtc_get(CRTHiOrd), 2);
	d0	= d0 << 2;

	return d0;
}
*/

/*
void SetBladeScreenStartAddress(DWORD d0)
{
	DWORD d1;

	d0 = d0 >> 2;
	card_crtc_set(0x0D, (BYTE)(d0 & 0xFF));
	card_crtc_set(0x0C, (BYTE)((d0 >> 8)  & 0xFF));

	d1 = card_crtc_get(CRTCModuleTest);
	d1 = card_bitmov(d1, 5, d0, 16);
	card_crtc_set(CRTCModuleTest, (BYTE)d1);

	d1 = card_crtc_get(CRTHiOrd);
	d1 = card_bitmov(d1, 0, d0, 17);
	d1 = card_bitmov(d1, 1, d0, 18);
	d1 = card_bitmov(d1, 2, d0, 19);
	card_crtc_set(CRTHiOrd, (BYTE)d1);
}
*/

static void blade3d_set_hsync_skew_in_misc(int flag) {
	DWORD d0;

	if (flag == 0) {
		flag = 1;
	} else {
		flag = 0;		// no skew
	}
	d0 = card_graph_get(MiscInternalControlReg);
	d0 = card_bitmov(d0, 5, flag, 0);
	card_graph_set(MiscInternalControlReg, (BYTE)d0);
}

/*
static int GetDoubleLogicalLineWidth(void)
{
	DWORD d0;

	d0 = card_graph_get(MiscInternalControlReg);
	d0 = d0 >> 2;
	d0 = d0 & 0x01;

	return (int)d0;
}
*/

const char* blade3d_driver(void) {
	return blade3d_card->name;
}

int blade3d_detect(void) {
	int i;

#if 0 /* fake detect */
	blade3d_card = blade3d_id_list + 0;
	return 1;
#endif

	if (pci_detect()!=0) {
		CARD_LOG(("blade3d: PCI BIOS not installed.\n"));
		return 0;
	}

	for(i=0;blade3d_id_list[i].name;++i) {
		if (pci_find_device(0x00001023,blade3d_id_list[i].value,&blade3d_bus_device_func)==0)
			break;
	}

	if (!blade3d_id_list[i].name) {
		CARD_LOG(("blade3d: not found\n"));
		return 0;
	}

	CARD_LOG(("blade3d: found %s, device id %04x\n", blade3d_id_list[i].name, blade3d_id_list[i].value));

	if (!blade3d_id_list[i].supported) {
		CARD_LOG(("blade3d: card not supported\n"));
		return 0;
	}

	blade3d_card = blade3d_id_list + i;

	return 1;
}

void blade3d_reset(void) {
}

int blade3d_set(const card_crtc STACK_PTR* ccp) {
	card_crtc cp = *ccp;
	int pll_mul, pll_div, pll_p, pll_s;

	blade3d_new_mode();
	blade3d_unprotect_registers();

	cp.dotclockHz = blade3d_clock_compute(cp.dotclockHz, &pll_mul, &pll_div, &pll_p, &pll_s);
	if (pll_mul==0 || cp.dotclockHz<0) {
		CARD_LOG(("blade3d: unable to compute programmable clock\n"));
		return 0;
	}

	cp.HBStart = cp.HDisp;
	cp.HBEnd = cp.HSEnd;
	cp.VBStart = cp.VDisp;
	cp.VBEnd = cp.VSEnd;

	card_signal_disable();
	card_generic_all_set(&cp);
	blade3d_ext_set(&cp);
	blade3d_set_hsync_skew_in_misc(0);
	blade3d_clock_set(pll_mul, pll_div, pll_p, pll_s);
	card_signal_enable();

	return 1;
}

