/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 *
 * This video driver is STRONGLY derived from the video drivers of the
 * VSyncMAME project by S.Sakamaki (Saka)
 */

#include "compil.h"
#include "s6326.h"
#include "card.h"

#define SiS6326    1
#define SiS620     2
#define SiS530     3
#define SiS300     4
#define SiS630     5
#define SiS540     6

static void SetSiS6326ExtRegister(card_crtc STACK_PTR *cp);
static int GetSiS6326BytePerLine(void);
static void SetSiS6326BytePerLine(int d0);
static void SetSiS6326Interlace(int flag);
static int SetSiS6326VClock(BYTE Numerator, BYTE Divider, BYTE DeNumerator, BYTE PostScalar, BYTE SeqDivider);
static void SetSiS300ExtRegister(card_crtc STACK_PTR *cp);
static int GetSiS300BytePerLine(void);
static void SetSiS300BytePerLine(int d0);
static void SetSiS300Interlace(int flag);
static int SetSiS300VClock(BYTE Numerator, BYTE Divider, BYTE DeNumerator, BYTE PostScalar, BYTE SeqDivider);
static void SetSiS6326DoubleScan(int flag);
static void SetSis6326ClockRegisterInternal(void);
static int ExtSis6326RegisterUnLock(void);
static long GetSiS6326VclkParam(long vclk, int STACK_PTR *Num, int STACK_PTR *DeN, int STACK_PTR *Div, int STACK_PTR *PS, int STACK_PTR *SeqDiv);

static unsigned SiS6326BusDeviceFunc;
static DWORD Sis6326DeviceType;

const char* driverSiS6326(void) {
	switch (Sis6326DeviceType) {
		case SiS6326 : return "Sis6326";
		case SiS620 : return "Sis620";
		case SiS530 : return "Sis530";
		case SiS300 : return "Sis300";
		case SiS630 : return "Sis630";
		case SiS540 : return "Sis540";
	}
	return 0;
}

int detectSiS6326(void)
{
	if (pci_detect()!=0) {
		CARD_LOG(("matrox: PCI BIOS not Installed.\n"));
		return 0;
	}

	while(1){
		if (pci_find_device(0x00001039,0x00000620,&SiS6326BusDeviceFunc)==0) {
			CARD_LOG(("sis: SiS620 found\n"));
			Sis6326DeviceType = SiS620;
			break;
		}
		if (pci_find_device(0x00001039,0x00000530,&SiS6326BusDeviceFunc)==0) {
			CARD_LOG(("sis: SiS530 found\n"));
			Sis6326DeviceType = SiS530;
			break;
		}
		if (pci_find_device(0x00001039,0x00006326,&SiS6326BusDeviceFunc)==0) {
			CARD_LOG(("sis: SiS6326 found\n"));
			Sis6326DeviceType = SiS6326;
			break;
		}
		if (pci_find_device(0x00001039,0x00000300,&SiS6326BusDeviceFunc)==0) {
			CARD_LOG(("sis: SiS300 found\n"));
			Sis6326DeviceType = SiS300;
			break;
		}
		if (pci_find_device(0x00001039,0x00000630,&SiS6326BusDeviceFunc)==0) {
			CARD_LOG(("sis: SiS630 found\n"));
			Sis6326DeviceType = SiS630;
			break;
		}
		if (pci_find_device(0x00001039,0x00000540,&SiS6326BusDeviceFunc)==0) {
			CARD_LOG(("sis: SiS540 found\n"));
			Sis6326DeviceType = SiS540;
			break;
		}

		CARD_LOG(("sis: not found\n"));
		return 0;
	}


	if(ExtSis6326RegisterUnLock() == 0){
		CARD_LOG(("sis: Failed unlock extention registers\n"));
		return 0;
	}

	return 1;
}

static int ExtSis6326RegisterUnLock(void)
{
	BYTE d0;

	d0 = card_seq_get(0x05);
	if(d0 == 0xA1)
		return 1;
	card_seq_set(0x05, 0x86);
	if(card_seq_get(0x05) == 0xA1){
		CARD_LOG(("sis: SiS extention registers unlocked\n"));
		return 1;
	}
	card_seq_set(0x05, d0);
	return 0;
}


void resetSiS6326(void)
{
}

int setSiS6326(const card_crtc STACK_PTR* ccp)
{
	card_crtc cp = *ccp;
	int Num,DeN,Div,PS,SeqDiv;

	cp.dotclockHz = GetSiS6326VclkParam(cp.dotclockHz, &Num, &Div, &DeN, &PS, &SeqDiv);

	cp.HBStart = cp.HDisp;
	cp.HBEnd   = cp.HTotal - 8;
	cp.VBStart = cp.VDisp;
	cp.VBEnd   = cp.VTotal - 1;

	card_signal_disable();
	switch(Sis6326DeviceType){
		case SiS6326:
		case SiS620:
		case SiS530:
			SetSiS6326VClock(Num, Div, DeN, PS, SeqDiv);
			card_generic_all_set(&cp);
			SetSiS6326ExtRegister(&cp);
			SetSiS6326DoubleScan(cp.doublescan);
			SetSiS6326Interlace(cp.interlace);
			card_polarity_set(cp.hpolarity, cp.vpolarity);
			break;

		case SiS300:
		case SiS630:
		case SiS540:
			SetSiS300VClock(Num, Div, DeN, PS, SeqDiv);
			card_generic_all_set(&cp);
			SetSiS300ExtRegister(&cp);
			SetSiS6326DoubleScan(cp.doublescan);
			SetSiS300Interlace(cp.interlace);
			card_polarity_set(cp.hpolarity, cp.vpolarity);
			break;
	}

	card_signal_enable();
	return 1;
}

static void SetSis6326ClockRegisterInternal(void)
{
	BYTE d0;

	d0 = card_in(0x3CC);
	d0 = d0 | 0x0C;
	card_out(0x3C2, d0);
}


static BYTE PostScalarTable[]={ 0, 0x00, 0x01, 0x02, 0x03, 0, 0x02, 0, 0x03};

static int SetSiS6326VClock(BYTE Numerator, BYTE Divider, BYTE DeNumerator, BYTE PostScalar, BYTE SeqDivider)
{
	BYTE _2B_DeN;
	BYTE _2B_PS;
	BYTE _2A_Div;
	BYTE _2A_Num;
	BYTE _13_PS;
	BYTE d0;
	int d1;

	SetSis6326ClockRegisterInternal();
	card_seq_set(0x38, card_seq_get(0x38) & 0xFC);

	d0 = 0x01;
	if(SeqDivider == 2)  d0 = d0 | 0x08;
	card_seq_set(0x01, d0);

	_2A_Div = ((Divider == 1) ? 0 : 1) << 7;
	_2A_Num = (Numerator - 1) & 0x7F;
	_13_PS  = ((PostScalar == 6 || PostScalar == 8) ? 1 : 0) << 6;
	_2B_PS  = PostScalarTable[PostScalar] << 5;
	_2B_DeN = (DeNumerator - 1) & 0x1F;

	d0 = _2A_Div | _2A_Num;
	card_seq_set(0x2A, d0);
	d0 = card_seq_get(0x13);
	d0 = d0 & ~0x40;
	d0 = d0 | _13_PS;
	card_seq_set(0x13, d0);
	d0 = _2B_PS | _2B_DeN;
//      if( frequency >= high ) d0 |= 0x80;
	card_seq_set(0x2B, d0);                       // trigger ?

	d1 = 65536;
	while(d1 > 0){
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		d1 -= 1;
	}

	return 0;
}

static int SetSiS300VClock(BYTE Numerator, BYTE Divider, BYTE DeNumerator, BYTE PostScalar, BYTE SeqDivider)
{
	BYTE d0;
	int d1;
	BYTE xr2b;
	BYTE xr2c;
	int sbit;

	SetSis6326ClockRegisterInternal();

	d0 = 0x01;
	if(SeqDivider == 2)  d0 = d0 | 0x08;
	card_seq_set(0x01, d0);

	sbit = (PostScalar == 6 || PostScalar == 8) ? 1 : 0;
	PostScalar = PostScalarTable[PostScalar];

	xr2b = (Numerator - 1) & 0x7f;
	if (Divider == 2)
		xr2b |= 0x80;
	xr2c = (DeNumerator - 1) & 0x1f;
	xr2c |= PostScalar << 5;
	if (sbit)
		xr2c |= 0x80;
	card_seq_set(0x2B, xr2b);
	card_seq_set(0x2C, xr2c);
	card_seq_set(0x2D, 0x80);

	d1 = 65536;
	while(d1 > 0){
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		inportb(0x3da);
		d1 -= 1;
	}

	return 0;
}

static void SetSiS6326ExtRegister(card_crtc STACK_PTR *cp)
{
	int d0;

	d0 = card_seq_get(0x0A);
	d0 = card_bitmov(d0, 0, cp->VTotal  - 2,  10);
	d0 = card_bitmov(d0, 1, cp->VDisp   - 1 , 10);
	d0 = card_bitmov(d0, 2, cp->VBStart     , 10);
	d0 = card_bitmov(d0, 3, cp->VSStart     , 10);
	card_seq_set(0x0A, d0);


	d0 = card_bitmov(d0, 0, (cp->HTotal  >> 3) - 5,      8);
	d0 = card_bitmov(d0, 1, (cp->HDisp   >> 3) - 1,      8);
	d0 = card_bitmov(d0, 2, (cp->HBStart >> 3)    ,      8);
	d0 = card_bitmov(d0, 3, (cp->HSStart >> 3)    ,      8);
	d0 = card_bitmov(d0, 4, (cp->HBEnd   >> 3)    ,      6);
#if 1
	d0 = card_bitmov(d0, 5, cp->HSStart, 0);
	d0 = card_bitmov(d0, 6, cp->HSStart, 1);
	d0 = card_bitmov(d0, 7, cp->HSStart, 2);
#else
	d0 = d0 & 0x1F;
#endif
	card_seq_set(0x12, d0);

}

static void SetSiS300ExtRegister(card_crtc STACK_PTR *cp)
{
	int d0;

	d0 = (
		(((cp->VSEnd)    & 0x10 ) <<  1) |               /* D5 */
		(((cp->VBEnd)    & 0x100) >>  4) |               /* D4 */
		(((cp->VSStart)  & 0x400) >>  7) |               /* D3 */
		(((cp->VBStart)  & 0x400) >>  8) |               /* D2 */
		(((cp->VDisp-1)  & 0x400) >>  9) |               /* D1 */
		(((cp->VTotal-2) & 0x400) >> 10));               /* D0 */
	card_seq_set(0x0A, d0);

	d0 = (
		(( (cp->HSStart >> 3)      & 0x300) >> 2)  |  /* D76 */
		(( (cp->HBStart >> 3)      & 0x300) >> 4)  |  /* D54 */
		((((cp->HDisp   >> 3) - 1) & 0x300) >> 6)  |  /* D32 */
		((((cp->HTotal  >> 3) - 5) & 0x300) >> 8));   /* D10 */
	card_seq_set(0x0B, d0);

	d0      = card_seq_get(0x0C) & 0xF8;
	d0 |= (((cp->HSEnd >> 3) & 0x20) >> 3) |          /* D2 */
		(((cp->HBEnd >> 3) & 0xC0) >> 6);       /* D10 */
	card_seq_set(0x0C, d0);

}

static int GetSiS6326BytePerLine(void)
{
	DWORD d0;

	d0      = (DWORD)card_crtc_get(0x13);
	d0 |= (DWORD)((card_seq_get(0x0A) << 4) & 0x0F00);
	d0 = d0 << 3;
	return d0;
}

static void SetSiS6326BytePerLine(int d0)
{
	BYTE d1;

	d0 = d0 >> 3;
	d1 = (BYTE)(d0 & 0xFF);
	card_crtc_set(0x13, d1);
	d1 = card_seq_get(0x0A) & 0x0F;
	d1 = d1 | (BYTE)((d0 >> 4) & 0xF0);
	card_seq_set(0x0A, d1);
}

static void SetSiS6326DoubleScan(int flag)
{
	if(flag == 0){
		card_crtc_set(0x09, card_bitmov(card_crtc_get(0x09), 7, 0, 0));
	}else{
		card_crtc_set(0x09, card_bitmov(card_crtc_get(0x09), 7, 1, 0));
	}
}

static void SetSiS6326Interlace(int flag)
{
	BYTE d1;

	if(flag == 0){
		d1 = card_seq_get(0x06);
		if( (d1 & (1 << 5)) != 0){
			d1 = d1 & ~(1 << 5);                                    // Interlace Disable
			card_seq_set(0x06, d1);
			SetSiS6326BytePerLine( GetSiS6326BytePerLine() / 2);
		}
	}else{
		d1 = card_seq_get(0x06);
		if( (d1 & (1 << 5)) == 0){
			d1 = d1 | (1 << 5);                                     // Interlace Enable
			card_seq_set(0x06, d1);
			SetSiS6326BytePerLine( GetSiS6326BytePerLine() * 2);
		}
	}
}

static int GetSiS300BytePerLine(void)
{
	DWORD d0;

	d0      = (DWORD)card_crtc_get(0x13) & 0xFF;
	d0 |= (DWORD)(card_seq_get(0x0E) << 8) & 0x0F00;
	d0 = d0 << 3;
	return d0;
}

static void SetSiS300BytePerLine(int d0)
{
	BYTE d1;

	d0 = d0 >> 3;
	d1 = (BYTE)(d0 & 0xFF);
	card_crtc_set(0x13, d1);
	d1 = card_seq_get(0x0E) & 0xF0;
	d1 = d1 | ((d0 >> 8) & 0x0F);
	card_seq_set(0x0E, d1);
}

static void SetSiS300Interlace(int flag)
{
	BYTE d1;

	if(flag == 0){
		d1 = card_seq_get(0x06);
		if( (d1 & (1 << 5)) != 0){
			d1 = d1 & ~(1 << 5);                                    // Interlace Disable
			card_seq_set(0x06, d1);
			SetSiS300BytePerLine( GetSiS300BytePerLine() / 2);
		}
	}else{
		d1 = card_seq_get(0x06);
		if( (d1 & (1 << 5)) == 0){
			d1 = d1 | (1 << 5);                                     // Interlace Enable
			card_seq_set(0x06, d1);
			SetSiS300BytePerLine( GetSiS300BytePerLine() * 2);
		}
	}
}

#if 1

#define MAX_VCO 135000000
#define BASE_FREQ 14318000

static long GetSiS6326VclkParam(long vclk, int STACK_PTR *Num, int STACK_PTR *Div, int STACK_PTR *DeN, int STACK_PTR *PS, int STACK_PTR *SeqDiv)
{
	double desfreq;
	double temp;
	double tempdt;
	double dt;
	int d0;
	int d1;
	int P[2];
	int PD = 0;
	int idx;

	if(vclk > 135000000) vclk = 135000000;

	P[0] = 1;
	P[1] = 1;
	if(vclk < (BASE_FREQ * 6) / 2){
		P[0] = 2;
		P[1] = 3;
	}
	if(vclk < (BASE_FREQ * 6) / 3){
		P[0] = 3;
		P[1] = 4;
	}
	if(vclk < (BASE_FREQ * 6) / 4){
		P[0] = 4;
		P[1] = 6;
	}
	if(vclk < (BASE_FREQ * 6) / 6){
		P[0] = 6;
		P[1] = 8;
	}
	if(vclk < (BASE_FREQ * 6) / 8){
		P[0] = 8;
		P[1] = 12;
	}
	if(vclk < (BASE_FREQ * 6) / 12){
		P[0] = 12;
		P[1] = 16;
	}
	if(vclk < (BASE_FREQ * 6) / 16){
		P[0] = 16;
		P[1] = 16;
	}

	*Num = 0;
	*DeN = 0;
	dt = MAX_VCO;
	idx = 0;
	while(idx < 2){
		desfreq = vclk;
		for(d0 = 2; d0 <= 128*2; d0 += 1){
			if((128 <= d0) && (d0 % 2 == 1))
				continue;
			for(d1 = 3; d1 <= 32; d1 += 1){
				temp = BASE_FREQ * (double)d0 / d1;
				if(temp > MAX_VCO)
					continue;
				temp = temp / P[idx];
				if(desfreq > temp)
					tempdt = desfreq - temp;
				else
					tempdt = temp - desfreq;
				if(tempdt < dt){
					*Num = d0;
					*DeN = d1;
					dt = tempdt;
					PD = P[idx];
				}
			}
		}
		idx += 1;
	}

	if(*Num > 128){
		*Div  = 2;
		*Num /= 2;
	}else{
		*Div  = 1;
	}

	if(PD <= 8){
		*SeqDiv = 1;
		*PS = PD;
	}else{
		*SeqDiv = 2;
		*PS = PD / 2;
	}


	vclk = (long)((double)BASE_FREQ * (*Num * *Div)/(*DeN * *PS * *SeqDiv));

	CARD_LOG(("sis: Num %3d, DeN %3d, Div %1d, PS %1d, SeqDiv %1d\n",
				*Num, *DeN, *Div, *PS, *SeqDiv));
	CARD_LOG(("sis: 14.318 * (Num/DeN) : %lf\n", (BASE_FREQ/1000000)*(*Num)/(*DeN)));
	CARD_LOG(("sis: vclk : %lf\n", (double)vclk / 1000000.0));

	return vclk;
}


#else





//#define MAX_VCO 353000000
#define MAX_VCO 135000000
#define BASE_FREQ 14318000

static long GetVclkParam(long vclk, int *Num, int *Div,
							int *DeN, int *PS, int *SeqDiv)
{
	double desfreq;
	double temp;
	double tempdt;
	double dt;
	int d0;
	int d1;
	int P[2];
	int PD = 0;
	int idx;

	if(vclk > 135000000) vclk = 135000000;

	P[0] = 1;
	P[1] = 1;
	if(vclk < (BASE_FREQ * 6) / 2){
		P[0] = 2;
		P[1] = 3;
	}
	if(vclk < (BASE_FREQ * 6) / 3){
		P[0] = 3;
		P[1] = 4;
	}
	if(vclk < (BASE_FREQ * 6) / 4){
		P[0] = 4;
		P[1] = 6;
	}
	if(vclk < (BASE_FREQ * 6) / 6){
		P[0] = 6;
		P[1] = 8;
	}
	if(vclk < (BASE_FREQ * 6) / 8){
		P[0] = 8;
		P[1] = 12;
	}
	if(vclk < (BASE_FREQ * 6) / 12){
		P[0] = 12;
		P[1] = 16;
	}
	if(vclk < (BASE_FREQ * 6) / 16){
		P[0] = 16;
		P[1] = 16;
	}

	*Num = 0;
	*DeN = 0;
	dt = MAX_VCO;
	idx = 0;
	while(idx < 2){
		desfreq = vclk;
		for(d0 = 2; d0 <= 128*2; d0 += 1){
			if((128 <= d0) && (d0 % 2 == 1))
				continue;
			for(d1 = 3; d1 <= 32; d1 += 1){
				temp = BASE_FREQ * (double)d0 / d1;
				if(temp > MAX_VCO)
					continue;
				temp = temp / P[idx];
				if(desfreq > temp)
					tempdt = desfreq - temp;
				else
					tempdt = temp - desfreq;
				if(tempdt < dt){
					*Num = d0;
					*DeN = d1;
					dt = tempdt;
					PD = P[idx];
				}
			}
		}
		idx += 1;
	}

	if(*Num > 128){
		*Div  = 2;
		*Num /= 2;
	}else{
		*Div  = 1;
	}

	if(PD <= 8){
		*SeqDiv = 1;
		*PS = PD;
	}else{
		*SeqDiv = 2;
		*PS = PD / 2;
	}


	vclk = (long)((double)BASE_FREQ * (*Num * *Div)/(*DeN * *PS * *SeqDiv));

	logerror("Num %3d, DeN %3d, Div %1d, PS %1d, SeqDiv %1d\n",
				*Num, *DeN, *Div, *PS, *SeqDiv);
	logerror("14.318 * (Num/DeN) : %lf\n", (BASE_FREQ/1000000)*(*Num)/(*DeN));
	logerror("vclk : %lf\n", (double)vclk / 1000000.0);

	return vclk;
}


#endif
