#include "dirty.h"

#include <assert.h>

_dirty_t _dirty[_DIRTY_DX*_DIRTY_DY];

/* Number of dirty block */
unsigned dirty_dx;
unsigned dirty_dy;

/* Mask for back dependency */
static UINT32 dirty_mask;

/* Inizialize the dirty system
	level number of dirty levels
	dx,dy number of dirty blocks
*/
void dirty_init(unsigned level, unsigned dx, unsigned dy) {
	unsigned mask;

	assert( dirty_dx == 0 && dirty_dy == 0);
	assert( level < 8 ); /* if ==8 is not possible to shift 4 byte at one time */
	assert( dx <= _DIRTY_DX );
	assert( dy <= _DIRTY_DY );

	dirty_dx = dx;
	dirty_dy = dy;

	mask = (1 << level) - 1;
	dirty_mask = mask | (mask << 8) | (mask << 16) | (mask << 24);

	/* all is dirty */
	memset(_dirty,mask,sizeof(_dirty));
}

/* Deinizialize the dirty system */
void dirty_done(void) {
	assert( dirty_dx != 0 && dirty_dy != 0 );
#ifndef NDEBUG
	dirty_dx = 0;
	dirty_dy = 0;
	dirty_mask = 0;
#endif
}

/* Shift dirty table of one position */
void dirty_shift(void) {
	unsigned y;
	unsigned q_dx = (dirty_dx + 3) / 4; /* possible overrun */
	for(y=0;y<dirty_dy;++y) {
		UINT32* p = (UINT32*)(_dirty + y * _DIRTY_DX);
		unsigned counter = q_dx;
		while (counter) {
			*p = (*p << 1) & dirty_mask;
			++p;
			--counter;
		}
	}
}

void dirty_set_region(unsigned x, unsigned y, unsigned dx, unsigned dy) {
	unsigned iy;
	assert( x + dx <= dirty_dx && y + dy <= dirty_dy );
	for(iy=0;iy<dy;++iy) {
		UINT8* p = _dirty + (iy+y) * _DIRTY_DX + x;
		unsigned counter = dx;
		while (counter && (((unsigned)p) & 0x3)!=0) {
			*p |= 0x01;
			++p;
			--counter;
		}
		while (counter>=4) {
			*((UINT32*)p) |= 0x01010101;
			p += 4;
			counter -= 4;
		}
		while (counter) {
			*p |= 0x01;
			++p;
			--counter;
		}
	}
}

void dirty_set_all(void) {
	dirty_set_region(0,0,dirty_dx,dirty_dy);
}

void dirty_set_bitmap_region(unsigned x1, unsigned y1, unsigned x2, unsigned y2) {
	unsigned bx1 = x1/DIRTY_DELTA_X;
	unsigned by1 = y1/DIRTY_DELTA_Y;
	unsigned bx2 = x2/DIRTY_DELTA_X;
	unsigned by2 = y2/DIRTY_DELTA_Y;
	dirty_set_region(bx1,by1,bx2-bx1+1,by2-by1+1);
}

void dirty_set_bitmap_region_clip(int x1, int y1, int x2, int y2) {
	int bx1 = x1/DIRTY_DELTA_X;
	int by1 = y1/DIRTY_DELTA_Y;
	int bx2 = x2/DIRTY_DELTA_X;
	int by2 = y2/DIRTY_DELTA_Y;
	int bdx,bdy;
	if (bx1<0)
		bx1 = 0;
	if (by1<0)
		by1 = 0;
	if (bx2>=dirty_dx)
		bx2 = dirty_dx-1;
	if (by2>=dirty_dy)
		by2 = dirty_dy-1;
	bdx = bx2-bx1+1;
	bdy = by2-by1+1;
	if (bdx>0 && bdy>0)
		dirty_set_region(bx1,by1,bdx,bdy);
}

