#ifndef __DIRTY_H
#define __DIRTY_H

#include "osd_cpu.h"

/* Size of one dirty block */
#define DIRTY_DELTA_X 16
#define DIRTY_DELTA_Y 16

/* Private */

typedef UINT8 _dirty_t;

#define _DIRTY_DX 256
#define _DIRTY_DY 256

extern _dirty_t _dirty[_DIRTY_DX*_DIRTY_DY];

/* Number of dirty block */
unsigned dirty_dx;
unsigned dirty_dy;

/* Public */

/* Mark one block dirty
	x,y block coordinate (== bitmap coordinate/DIRTY_DELTA_XY )
*/
INLINE void dirty_set(unsigned _x, unsigned _y) {
	_dirty[_x+_y*_DIRTY_DX] |= 0x01;
}

/* Check if one block is dirty
	x,y block coordinate (== bitmap coordinate/DIRTY_DELTA_XY )
*/
INLINE _dirty_t is_dirty(unsigned _x, unsigned _y) {
	return _dirty[_x+_y*_DIRTY_DX];
}

void dirty_init(unsigned _level, unsigned _dx, unsigned _dy);
void dirty_done(void);
void dirty_shift(void);

/* Set dirty one block region
	x,y start dirty block
	dx,dy number of dirty block
*/
void dirty_set_region(unsigned _x, unsigned _y, unsigned _dx, unsigned _dy);

/* Set all dirty */
void dirty_set_all(void);

/* Set dirty one bitmap region
	x1,y1 upper left corner	(included)
	x2,y2 down right corner (included)
*/
void dirty_set_bitmap_region(unsigned _x1, unsigned _y1, unsigned _x2, unsigned _y2);

/* Set dirty one bitmap region with clipping
	x1,y1 upper left corner	(included)
	x2,y2 down right corner (included)
*/
void dirty_set_bitmap_region_clip(int _x1, int _y1, int _x2, int _y2);


#endif
