#include "record.h"

#include <stdio.h>
#include <stdlib.h>

// ---------------------------------------------------------------------------
// variables

int sound_record_time; /* seconds to record, ==0 disable recording */

// ---------------------------------------------------------------------------
// fread/fwrite

size_t fread_u16(unsigned* p, size_t num, FILE* f) {
	unsigned count = 0;
	while (count<num) {
		unsigned char b0,b1;
		if (fread(&b0, 1, 1, f ) != 1 || fread(&b1, 1, 1, f ) != 1) 
			return count;
		*p = b0 | (unsigned)b1 << 8;
		++count;
		++p;
	}
	return count;
}

size_t fwrite_u16(unsigned* p, size_t num, FILE* f) {
	unsigned count = 0;
	while (count<num) {
		unsigned char b0,b1;
		b0 = *p & 0xFF;
		b1 = (*p >> 8) & 0xFF;
		if (fwrite(&b0, 1, 1, f) != 1 || fwrite(&b1, 1, 1, f) != 1) 
			return count;
		++count;
		++p;
	}
	return count;
}

size_t fread_u32(unsigned* p, size_t num, FILE* f) {
	unsigned count = 0;
	while (count<num) {
		unsigned char b0,b1,b2,b3;
		if (fread( &b0, 1, 1, f ) != 1 || fread( &b1, 1, 1, f ) != 1 ||
			fread( &b2, 1, 1, f ) != 1 || fread( &b3, 1, 1, f ) != 1)
			return count;
		*p = b0 | (unsigned)b1 << 8 | (unsigned)b2 << 16 | (unsigned)b3 << 24;
		++count;
		++p;
	}
	return count;
}

size_t fwrite_u32(unsigned* p, size_t num, FILE* f) {
	unsigned count = 0;
	while (count<num) {
		unsigned char b0,b1,b2,b3;
		b0 = *p & 0xFF;
		b1 = (*p >> 8) & 0xFF;
		b2 = (*p >> 16) & 0xFF;
		b3 = (*p >> 24) & 0xFF;
		if (fwrite(&b0, 1, 1, f) != 1 || fwrite(&b1, 1, 1, f) != 1 ||
			fwrite(&b2, 1, 1, f) != 1 || fwrite(&b3, 1, 1, f) != 1)
			return count;
		++count;
		++p;
	}
	return count;
}

// ---------------------------------------------------------------------------
// wave

size_t fread_tag(char* p, FILE* f) {
	return fread( p, 1, 4, f ) == 4;
}

size_t fwrite_tag(char* p, FILE* f) {
	return fwrite( p, 1, 4, f ) == 4;
}

size_t fskip(size_t num, FILE* f) {    
	unsigned count = 0;
	while (count<num) {
		int c = getc(f);
		if (c==EOF) return count;
		++count;
	}
	return count;
}

// record_id

typedef struct {
		char id[4];
		unsigned size;
	} record_id;

size_t fread_id( record_id* p, FILE* f) {
	return fread_tag( p->id, f ) && fread_u32( &p->size, 1, f );
}

size_t fwrite_id( record_id* p, FILE* f) {
	return fwrite_tag( p->id, f ) && fwrite_u32( &p->size, 1, f );
}

#define WAVE_FORMAT_PCM      0x0001 // Microsoft Pulse Code Modulation (PCM) format
#define IBM_FORMAT_MULAW     0x0101 // IBM mu-law format
#define IBM_FORMAT_ALAW      0x0102 // IBM a-law format
#define IBM_FORMAT_ADPCM     0x0103 // IBM AVC Adaptive Differential Pulse Code Modulation format

typedef struct {
	unsigned wFormatTag;              // Format category
	unsigned wChannels;               // Number of channels
	unsigned dwSamplesPerSec;        // Sampling rate
	unsigned dwAvgu8sPerSec;       // For buffer estimation
	unsigned wBlockAlign;             // Data block size
} record_fmt;

size_t fread_fmt( record_fmt* p, FILE* f) {
	return fread_u16( &p->wFormatTag, 1, f ) &&
		fread_u16( &p->wChannels, 1, f ) &&
		fread_u32( &p->dwSamplesPerSec, 1, f ) &&
		fread_u32( &p->dwAvgu8sPerSec, 1, f ) &&
		fread_u16( &p->wBlockAlign, 1, f );
}

size_t fwrite_fmt( record_fmt* p, FILE* f) {
	return fwrite_u16( &p->wFormatTag, 1, f ) &&
		fwrite_u16( &p->wChannels, 1, f ) &&
		fwrite_u32( &p->dwSamplesPerSec, 1, f ) &&
		fwrite_u32( &p->dwAvgu8sPerSec, 1, f ) &&
		fwrite_u16( &p->wBlockAlign, 1, f );
}

// record_fmt_PCM

typedef struct {
	unsigned wBitsPerSample;      // Sample size
} record_fmt_specific_PCM;

size_t fread_fmt_PCM( record_fmt_specific_PCM* p, FILE* f) {
	return fread_u16( &p->wBitsPerSample, 1, f );
}

size_t fwrite_fmt_PCM( record_fmt_specific_PCM* p, FILE* f) {
	return fwrite_u16( &p->wBitsPerSample, 1, f);
}

size_t fwrite_header(int speed, int bit, int channel, int size, FILE* f) {
	record_id riff_id;
	char wave_id[4];
	record_id fmt_id;
	record_fmt fmt;
	record_fmt_specific_PCM fmt_PCM;
	record_id data_id;
	
	unsigned size_byte;
	if (bit <= 8)
		size_byte = 1;
	else if (bit <= 16)
		size_byte = 2;
	else
		size_byte = 4;
	
	memcpy(riff_id.id,"RIFF",4);
	riff_id.size = 0x24 + size;
	if (!fwrite_id( &riff_id, f )) return 0;
	
	memcpy(wave_id,"WAVE",4);
	if (!fwrite_tag( wave_id, f )) return 0;

	memcpy(fmt_id.id, "fmt ", 4);
	fmt_id.size = 0x10;
	if (!fwrite_id( &fmt_id, f )) return 0;
	
	fmt.wFormatTag = WAVE_FORMAT_PCM;
	fmt.wChannels = channel; 
	fmt.dwSamplesPerSec = speed;
	fmt.dwAvgu8sPerSec = size_byte*channel*speed;
	fmt.wBlockAlign = size_byte*channel;
	if (!fwrite_fmt( &fmt, f )) return 0;

	fmt_PCM.wBitsPerSample = bit;
	if (!fwrite_fmt_PCM( &fmt_PCM, f )) return 0;
	
	memcpy( data_id.id, "data", 4);
	data_id.size = size;
	if (!fwrite_id( &data_id, f )) return 0;

	return 1;
}

size_t fwrite_header_size(int size, FILE* f) {
	unsigned dsize;

	dsize = 0x24 + size;
	if (fseek(f,4,SEEK_SET)!=0) 
		return 0;
	if (!fwrite_u32( &dsize, 1, f)) 
		return 0;

	dsize = size;
	if (fseek(f,0x28,SEEK_SET)!=0) 
		return 0;
	if (!fwrite_u32( &dsize, 1, f)) 
		return 0;

	return 1;
}

size_t fread_header(int* speed, int* bit, int* channel, int* size, FILE* f) {
	record_id riff_id;
	char wave_id[4];
	record_id fmt_id;
	record_fmt fmt;
	record_fmt_specific_PCM fmt_PCM;
	record_id data_id;

	if (!fread_id( &riff_id, f )) return 0;
	if (memcmp( riff_id.id, "RIFF", 4) != 0) return 0;

	if (!fread_tag( wave_id, f )) return 0;
	if (memcmp( wave_id, "WAVE", 4) != 0) return 0;

	// read until get fmt tag
	if (!fread_id( &fmt_id, f )) return 0;
	while (!feof(f) && memcmp( fmt_id.id, "fmt ", 4)!=0) {
		// salta al prossimo chunk
		if (!fskip( fmt_id.size, f )) return 0;
		// legge il chunk
		if (!fread_id( &fmt_id, f )) return 0;
	}
	if (memcmp( fmt_id.id, "fmt ", 4) != 0) return 0;
	if (fmt_id.size < 16) return 0;
	
	if (!fread_fmt( &fmt, f )) return 0;
	if (fmt.wFormatTag != WAVE_FORMAT_PCM) return 0;
	if (fmt.wChannels < 1 || fmt.wChannels > 2) return 0;

	if (!fread_fmt_PCM( &fmt_PCM, f )) return 0;
	if (fmt_PCM.wBitsPerSample < 4 || fmt_PCM.wBitsPerSample>16) return 0;

	// skip fmt unknow extension
	if (fmt_id.size > 16) {
		if (!fskip( fmt_id.size - 16, f )) return 0; 
	}

	// read until get data tag
	if (!fread_id( &data_id, f )) return 0;
	while (!feof(f) && memcmp( data_id.id, "data", 4)!=0) {
		// salta al prossimo chunk
		if (!fskip( data_id.size, f )) return 0;
		// legge il chunk
		if (!fread_id( &data_id, f )) return 0;
	}
	if (memcmp( data_id.id, "data", 4) != 0) return 0;

	*speed = fmt.dwSamplesPerSec;
	*bit = fmt_PCM.wBitsPerSample;
	*channel = fmt.wChannels;
	*size = data_id.size;

	return 1;
}

// ---------------------------------------------------------------------------
// sound_record

static int sound_record_active;
static int sound_record_stereo; // 1 (==0) or 2 (!=0) channels
static unsigned sound_record_sample_size; // size in byte of one sample
static unsigned sound_record_frequency; // frequency
static unsigned sound_record_mac; // samples saved
static unsigned sound_record_max; // samples to save
static unsigned char* sound_record_map;
static char sound_record_file[256];

#define SOUND_FREQUENCY_ERROR 0.01
unsigned SOUND_FREQUENCY_MAP[] = {
	48000, /* MPEG1 */
	44100, /* MPEG1 */
	32000, /* MPEG1 */
	24000, /* MPEG2 */
	22050, /* MPEG2 */
	16000, /* MPEG2 */
	12000, /* MPEG2.5 */
	11025, /* MPEG2.5 */
	8000, /* MPEG2.5 */
	0
};

/* Start a sound recording */
int sound_record_start(const char* file, unsigned samples, unsigned frequency, int stereo) {
	unsigned* f;
	if (sound_record_active)
		return -1;

	/* adjust the frequency to match the MP3 standard */	
	f = SOUND_FREQUENCY_MAP;
	while (*f) {
		double v = *f;
		v /= frequency;
		if (v > 1)
			v = 1 / v;
		v = 1 - v;			
		if (v<SOUND_FREQUENCY_ERROR) {
			sound_record_frequency = *f;
			break;
		}
                 ++f;
	}
	if (!*f)
		sound_record_frequency = frequency;

	sound_record_stereo = stereo;
	sound_record_mac = 0;
	sound_record_max = samples;
	sound_record_sample_size = 2;
	if (stereo)
		sound_record_sample_size *= 2;
	sound_record_map = (unsigned char*)malloc(sound_record_max * sound_record_sample_size);
	if (!sound_record_map)
		return -1;
	strcpy(sound_record_file,file);

	sound_record_active = 1;

	return 0;
}

/* Insert some data in the sound recording. Automatically save if full
 *	map samples buffer
 *	mac number of 16 bit samples mono or stereo in little endian format
 */
int sound_record_pump(void* map, unsigned mac) {
	unsigned copy;

	if (!sound_record_active)
		return 0;

	copy = mac;
	if (sound_record_mac + copy > sound_record_max)
		copy = sound_record_max - sound_record_mac;

	if (copy) {
		memcpy(sound_record_map + sound_record_mac * sound_record_sample_size, map, copy * sound_record_sample_size);
		sound_record_mac += copy;
	}

	return 0;
}

/* Cancel a sound recording */
int sound_record_cancel(void) {
	/* not active */
	if (!sound_record_active)
		return -1;

	sound_record_active = 0;
	free(sound_record_map);
	return 0;
}

/* Save and stop the current file recording */
int sound_record_stop(unsigned* saved) {
	FILE* f;
	unsigned i;
	unsigned mono_sample;
	unsigned fade_sample;
	int v_min;
	double v_amp;

	/* not active */
	if (!sound_record_active)
		return -1;

	sound_record_active = 0;

	/* empty */
	if (!sound_record_max) {
		free(sound_record_map);
		return -1;
	}

	f = fopen(sound_record_file,"wb");
	if (!f) {
		free(sound_record_map);
		return -1;
	}

	if (fwrite_header(sound_record_frequency, 16, sound_record_stereo ? 2 : 1, sound_record_mac*sound_record_sample_size, f)!=1) {
		free(sound_record_map);
		fclose(f);
		remove(sound_record_file);
		return -1;
	}

	/* adjust the volume */
	v_min = 0;
	mono_sample = sound_record_mac;
	if (sound_record_stereo)
		mono_sample *= 2;
        fade_sample = sound_record_frequency / 10;
        if (sound_record_stereo)
		fade_sample *= 2;
	if (fade_sample > mono_sample/2)
		fade_sample = mono_sample/2;
	for(i=0;i<mono_sample;++i) {
		int v = sound_record_map[i*2] + (((int)(char)sound_record_map[i*2+1]) << 8);
		if (v > 0)
			v = -v;
		if (v < v_min)
			v_min = v;
	}
	v_amp = (double)-32767.0 / v_min;
	if (v_amp > 16.0)
		v_amp = 16.0; /* limit the factor */
	/* v_amp may be < 1.0 if the -32768 sample is present */
	if (v_amp < 1.0)
		v_amp = 1.0;
	for(i=0;i<mono_sample;++i) {
		int v = sound_record_map[i*2] + (((int)(char)sound_record_map[i*2+1]) << 8);
		double fa = v_amp;
		if (i<fade_sample)
			fa = fa * i / fade_sample;
		if (i>mono_sample-fade_sample)
			fa = fa * (mono_sample - i) / fade_sample;
		v = v * fa;
		sound_record_map[i*2] = v & 0xFF;
		sound_record_map[i*2+1] = (v >> 8) & 0xFF;
	}

	if (fwrite(sound_record_map,sound_record_mac * sound_record_sample_size,1,f)!=1) {
		free(sound_record_map);
		fclose(f);
		remove(sound_record_file);
		return -1;
	}

	free(sound_record_map);
	fclose(f);

	*saved = sound_record_mac;

	return 0;
}

