/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "scrclear.h"

#include <sys/movedata.h>
#include <sys/segments.h>
#include <sys/farptr.h>
#include <assert.h>
#include <stdlib.h>

/***************************************************************************/
/* Clear */

/* Clear screen
 * in:
 *   x,y destination pos
 *   src data source
 */
static void video_chained8_clear(unsigned x, unsigned y, unsigned dx, unsigned dy, uint8 src) {
	unsigned j;
	unsigned pre_count;
	unsigned post_count;
	unsigned inner_count;
	unsigned sel = video_write_selector();
	uint32 mask;

	src &= 0xFF;
	mask = src | (src << 8) | (src << 16) | (src << 24);

	pre_count = (4 - x) & 3;
	if (pre_count >= dx) pre_count = dx;
	inner_count = (dx - pre_count) / 4;
	post_count = dx - pre_count - inner_count * 4;

	_farsetsel(sel);

	for(j=0;j<dy;++j) {
		unsigned dst = video_write_line(y) + x;
		unsigned i;
		for(i=pre_count;i;--i) {
			_farnspokeb(dst, mask);
			dst += 1;
		}
		for(i=inner_count;i;--i) {
			_farnspokel(dst, mask);
			dst += 4;
		}
		for(i=post_count;i;--i) {
			_farnspokeb(dst, mask);
			dst += 1;
		}
		++y;
	}
}

static void video_chained16_clear(unsigned x, unsigned y, unsigned dx, unsigned dy, unsigned src) {
	unsigned j;
	unsigned pre_count;
	unsigned post_count;
	unsigned inner_count;
	unsigned sel = video_write_selector();
	uint32 mask;

	src &= 0xFFFF;
	mask = src | (src << 16);

	pre_count = (2 - x) & 1;
	if (pre_count >= dx) pre_count = dx;
	inner_count = (dx - pre_count) / 2;
	post_count = dx - pre_count - inner_count * 2;

	_farsetsel(sel);

	for(j=0;j<dy;++j) {
		unsigned dst = video_write_line(y) + 2*x;
		unsigned i;
		if (pre_count) {
			_farnspokeb(dst, mask);
			_farnspokeb(dst + 1, mask >> 8);
			dst += 2;
		}
		for(i=inner_count;i;--i) {
			_farnspokel(dst, mask);
			dst += 4;
		}
		if (post_count) {
			_farnspokeb(dst, mask);
			_farnspokeb(dst + 1, mask >> 8);
			/* dst += 2; */
		}
		++y;
	}
}

static void video_unchained8_clear(unsigned x, unsigned y, unsigned dx, unsigned dy, uint8 src) {
	unsigned p;
	unsigned sel = video_write_selector();
	uint32 mask;

	src &= 0xFF;
	mask = src | (src << 8) | (src << 16) | (src << 24);

	_farsetsel(sel);

	for(p=0;p<4;++p) {
		unsigned j;

		unsigned p_x;
		unsigned p_dx;
		unsigned pre_count;
		unsigned post_count;
		unsigned inner_count;

		p_x = x / 4;
		p_dx = (x + dx - 1) / 4 - p_x + 1;
		if ((x + dx - 1) % 4 < p)
			--p_dx;
		if (x % 4 > p) {
			--p_dx;
			++p_x;
		}

		pre_count = (4 - p_x) & 3;
		if (pre_count >= p_dx) pre_count = p_dx;
		inner_count = (p_dx - pre_count) / 4;
		post_count = p_dx - pre_count - inner_count * 4;

		video_unchained_plane_set(p);
		for(j=0;j<dy;++j) {
			unsigned dst = video_write_line(y+j) + p_x;
			unsigned i;
			for(i=pre_count;i;--i) {
				_farnspokeb(dst, mask);
				dst += 1;
			}
			for(i=inner_count;i;--i) {
				_farnspokel(dst, mask);
				dst += 4;
			}
			for(i=post_count;i;--i) {
				_farnspokeb(dst, mask);
				dst += 1;
			}
		}
	}
}

/* Clear the screen */
void video_clear(unsigned dst_x, unsigned dst_y, unsigned dst_dx, unsigned dst_dy, unsigned src) {
	switch (video_bytes_per_pixel()) {
		case 1 :
			if (video_unchained_active())
				video_unchained8_clear(dst_x,dst_y,dst_dx,dst_dy,src);
			else
				video_chained8_clear(dst_x,dst_y,dst_dx,dst_dy,src);
			break;
		case 2:
			video_chained16_clear(dst_x,dst_y,dst_dx,dst_dy,src);
			break;
	}
}
