#include "mamalleg.h"
#include "driver.h"
#include "mamedbg.h"
#include "video.h"
#include "record.h"
#include "ui_text.h"
#include "vidhrdw/vector.h"
#include "dirty.h"
#include "screen/screen.h"
#include "screen/scrblit.h"
#include "screen/vga.h"
#include "screen/vbe.h"
#include "screen/vgaline.h"

#include <pc.h>
#include <conio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/farptr.h>
#include <go32.h>
#include <math.h>
#include <limits.h>
#include "ticker.h"

/***************************************************************************/
/* Externs */

void scale_vectorgames(int gfx_width,int gfx_height,int *width,int *height);

/* in msdos/sound.c */
int msdos_update_audio(void);

/* in msdos/input.c */
void poll_joysticks(void);

/***************************************************************************/
/* Variables used for reading configuration parameters (see video.h) */

int video_vsync_request;
int video_vsync_adjust_request;
int wait_vsync_request;
int use_triplebuf_request;
int skiplines_request;
int skipcolumns_request;
char resolution_request[256];
int scanlines;
int stretch_request;
float osd_gamma_correction = 1.0;
int brightness;
float brightness_paused_adjust;
int blit_orientation = 0;
int blit_combine_request;
int dirty_request;
double turbo_speed;
double turbo_startup_speed;
int turbo_startup_time;

/***************************************************************************/
/* Internal state */

static int warming_up;

static video_monitor monitor;

static int video_fps,video_attributes,video_orientation;

/* Palette */
static int palette_total = 0; /* number of entry in the palette, ==0 if not already set */
static int palette_modifiable; /* bool, is osd_modify_pen callable ? */
static video_color* current_palette;
static int* palette_dirty_color;
static int palette_dirty;
static unsigned* palette_lookup; /* palette for non packed video mode */
static int palette_rgb; /* bool, a palette (hardware or software) is used */

/* Brightness */
static int bright_dirty;
static int bright_lookup[256];

static double vsync_frame_rate;

int frameskip,autoframeskip;
#define FRAMESKIP_LEVELS 12

/* All video modes */
static video_mode_container all_gfx_modes;

/* Blit pipeline */
struct video_pipeline_struct blit_pipeline;
int blit_pipeline_flag;
int blit_dp;
int blit_dw;
unsigned blit_x;
unsigned blit_y;
int blit_offset;
int blit_combine_effective; /* effective value used */

static int vector_game;
static int dirty_effective;
static int wait_vsync_effective;

int throttle = 1;       /* toggled by F10 */

static int gone_to_gfx_mode;
static int frameskip_counter;
static int frames_displayed;
static unsigned turbo_frame_counter; /* frame counter never reset during the game play */
static TICKER start_time,end_time;    /* to calculate fps average on exit */
#define FRAMES_TO_SKIP 20       /* skip the first few frames from the FPS calculation */
				/* to avoid counting the copyright and info screens */

/***************************************************************************/
/* Internal bitmap */

#if 0
static void internal_clear_bitmap(struct osd_bitmap *bitmap)
{
	int i;
	for (i = 0;i < bitmap->height;i++)
	{
		if (bitmap->depth == 16)
			memset(bitmap->line[i],0,2*bitmap->width);
		else
			memset(bitmap->line[i],0,bitmap->width);
	}
}
#endif

struct internal_bitmap {
	unsigned safety;
	unsigned char* data;
};

static struct osd_bitmap* internal_create_bitmap(int width, int height, int depth, int safety, unsigned align_shift)
{
	struct osd_bitmap *bitmap;
	int i;
	unsigned rows, row_bytes, safety_bytes, adjust_bytes;
	struct internal_bitmap* ib;

	bitmap = malloc(sizeof(struct osd_bitmap));
	if (!bitmap)
		return 0;

	bitmap->depth = depth;
	bitmap->width = width;
	bitmap->height = height;

	rows = height + 2 * safety;
	if (depth == 16) {
		safety_bytes = safety * 2;
		row_bytes = width * 2;
	} else {
		safety_bytes = safety;
		row_bytes = width;
	}
	safety_bytes = (safety_bytes + ALIGN_MASK) & ~ALIGN_MASK;
	row_bytes = (row_bytes + ALIGN_MASK) & ~ALIGN_MASK;
	row_bytes += 2 * safety_bytes;

	ib = malloc(sizeof(struct internal_bitmap));
	if (!ib) {
		free(bitmap);
		return 0;
	}

	ib->data = malloc(rows * row_bytes + ALIGN);
	if (!ib->data) {
		free(ib);
		free(bitmap);
		return 0;
	}
	adjust_bytes = ALIGN - ((((unsigned)ib->data) + align_shift) % ALIGN);

	/* clear ALL the bitmap, including the safety area */
	memset(ib->data, 0, rows * row_bytes + ALIGN);

	bitmap->line = malloc(rows * sizeof(unsigned char *));
	if (!bitmap->line) {
		free(ib->data);
		free(ib);
		free(bitmap);
		return 0;
	}

	ib->safety = safety;
	for (i = 0;i < rows;i++)
		bitmap->line[i] = ib->data + adjust_bytes + i * row_bytes + safety_bytes;
	bitmap->line += safety;
	bitmap->_private = ib;

	return bitmap;
}

static void internal_free_bitmap(struct osd_bitmap *bitmap)
{
	if (bitmap)
	{
		struct internal_bitmap* ib = bitmap->_private;
		unsigned safety = ib->safety;
		bitmap->line -= safety;
		free(bitmap->line);
		free(bitmap);
		free(ib->data);
		free(ib);
	}
}

/***************************************************************************/
/* OSD Bitmap */

/* Create a bitmap. */
/* VERY IMPORTANT: the function must allocate also a "safety area" 16 pixels wide all */
/* around the bitmap. This is required because, for performance reasons, some graphic */
/* routines don't clip at boundaries of the bitmap. */

struct osd_bitmap *osd_alloc_bitmap(int width,int height,int depth)
{
	if (depth != 8 && depth != 16)
		depth = 8;

	return internal_create_bitmap(width,height,depth,16,0);
}

#if 0 /* removed in MAME 0.37b8 */
/* Set the bitmap to black */
void osd_clearbitmap(struct osd_bitmap *bitmap)
{
	internal_clear_bitmap(bitmap);

	if (bitmap == Machine->scrbitmap || bitmap == artwork_real_scrbitmap)
	{
		extern int bitmap_dirty; /* in mame.c */
		if (dirty_effective)
			dirty_set_all();
		bitmap_dirty = 1;
	}
}
#endif

void osd_free_bitmap(struct osd_bitmap *bitmap)
{
	internal_free_bitmap(bitmap);
}

/***************************************************************************/
/*
	Description of the various coordinate systems and variables :

	game_[width,height]
		Size of the game screen (after any flipxy or swapxy).
	game_used_[width,height]
		Size of the used part of the game screen.
	game_used_[posx,posy]
		Position of the used part of the game screen.
	game_visible_[width,height] (old gfx_display_columns, gfx_display_lines)
		Size of the visible part of the game screen. Can be smaller
		than game_used_* if the video mode is too small. game_visible_
		is always internal of the game_used part.
	game_visible_[posx,posy] (old skipcolumns,skiplines)
		Position of the visible part in the game screen (not
		in the game_used part).
	screen_visible_*
		Part of the screen used for drawing. game_visible_* area is
		drawn in screen_visible_* area, eventually with stretch_request.
		screen_visible may be smaller than screen if game don't have
		the same aspect ratio of the screen.
	screen_stretch
		Type of stretching done (is different than stretch_request variable)
			SCREEN_NONE 1 multiplier
			SCREEN_INTEGER interger multiplier
			SCREEN_FRACTIONAL fractional multiplier
		This value is derivate from game_visible_size and screen_visible_size
	screen_stretch_factorx
	screen_stretch_factory
		Integer factor of expansion (not really used in fractional)
		A factor of 2 is intented as 1 ARCADE pixel result in 2 PC pixel.
		This value is derivate from game_visible_size and screen_visible_size
*/

/*
Aspetto:

pc_aspect_ratio
	Sono le le dimensioni dello schermo di un normale pc 4,3

arcade_aspect_ratio (non usato)
	Sono le le dimensioni dello schermo di un arcade 4,3
	se orizzontale e 3,4 se verticale

game_aspect_ratio
	E' la dimensione in pixel ARCADE di un quadrato (a vista)
	disegnato sullo schermo ARCADE alla particolare risoluzione ARCADE
	game_width,game_height.

	E' la combinazione di arcade_aspect_ratio con la particolare
	risoluzione game_width,game_height.

	Attualmente sono possibili solo 1/1 e 2/1 (Blasteroid)

	arcade_aspect_ratio_x   game_aspect_ratio_x   game_size_x
	--------------------- * ------------------- = -----------
	arcade_aspect_ratio_y   game_aspect_ratio_y   game_size_y

screen_aspect_ratio
	E' la dimensione in pixel PC di un quadrato (a vista)
	disegnato sullo schermo PC alla particolare risoluzione PC.
	mode->size_x,mode->size_y.

	E' la combinazione di pc_aspect_ratio con la particolare
	risoluzione mode->size_x,mode->size_y.

	pc_aspect_ratio_x   screen_aspect_ratio_x   screen_size_x
	----------------- * --------------------- = -------------
	pc_aspect_ratio_y   screen_aspect_ratio_y   screen_size_y


resulting_aspect_ratio
	E' la dimensione in pixel ARCADE di un quadrato (a vista)
	disegnato sullo schermo PC.

	Idealmente dovrebbe essere uguale a game_aspect_ratio

	screen_stretch_factor_x   resulting_aspect_ratio_x   screen_aspect_ratio_x
	----------------------- * ------------------------ = ---------------------
	screen_stretch_factor_y   resulting_aspect_ratio_y   screen_aspect_ratio_y

	screen_stretch_factor_x   pc_aspect_ratio_x   resulting_aspect_ratio_x   screen_size_x
	----------------------- * ----------------- * ------------------------ = -------------
	screen_stretch_factor_y   pc_aspect_ratio_y   resulting_aspect_ratio_y   screen_size_y
*/

/* Default monitor aspect ratio */
#define pc_aspect_ratio_x 4
#define pc_aspect_ratio_y 3

/* Game color depth */
static int game_bits_per_pixel;
static int game_bytes_per_pixel;

static int game_width;
static int game_height;
static unsigned game_used_width;
static unsigned game_used_height;
static unsigned game_used_posx;
static unsigned game_used_posy;
static unsigned game_aspect_ratio_x;
static unsigned game_aspect_ratio_y;
static unsigned game_visible_posx;
static unsigned game_visible_posy;

/* Basic increment of number of pixel for mantaining the alignement */
static unsigned game_visible_posx_increment;

/* Video configuration */
typedef struct video_config {
	video_mode* mode;
	unsigned game_visible_width;
	unsigned game_visible_height;
	unsigned screen_visible_width;
	unsigned screen_visible_height;
	unsigned result_aspect_ratio_x; /* derivate */
	unsigned result_aspect_ratio_y;	/* derivate */
} video_config;

/* Active graphics configuration */
static video_config* gfx_cfg = 0;

/* Active comparing table */
struct video_config_compare_factor_table gfx_cmp;

/* Possible graphics configurations */
#define GFX_CFG_MAX 2048
static video_config gfx_cfg_map[GFX_CFG_MAX];
static unsigned gfx_cfg_mac = 0;

#define SCREEN_STRETCH_NONE 0
#define SCREEN_STRETCH_INTEGER 1
#define SCREEN_STRETCH_FRACTIONAL 2

static unsigned screen_stretch;
static unsigned screen_stretch_factorx;
static unsigned screen_stretch_factory;

/* Compute the MCD of two number (Euclide) */
static unsigned aspect_MCD(unsigned m, unsigned n) {
	while (n) {
		unsigned r = m % n;
		m = n;
		n = r;
	}
	return m;
}

/* Reduce a fraction */
static void aspect_reduce(unsigned* a, unsigned *b) {
	unsigned r = aspect_MCD(*a,*b);
	*a /= r;
	*b /= r;
}

/* Return the name of a video configuration (uniq) */
static const char* video_configuration_name(const video_config* cfg) {
	static char buffer[256];
	sprintf(buffer,"%s-%dx%d-%dx%d", video_mode_name(cfg->mode), cfg->game_visible_width, cfg->game_visible_height, cfg->screen_visible_width, cfg->screen_visible_height);
	return buffer;
}

/* Return the stretch used by the video configuration */
static int video_configuration_stretch(const video_config* gfx) {
	if (gfx->screen_visible_width == gfx->game_visible_width
		&& gfx->screen_visible_height == gfx->game_visible_height)
		return STRETCH_NONE;
	if (gfx->screen_visible_width % gfx->game_visible_width == 0
		&& gfx->screen_visible_height % gfx->game_visible_height == 0)
		return STRETCH_INTEGER_XY;
	if (gfx->screen_visible_width % gfx->game_visible_width == 0)
		return STRETCH_INTEGER_X_FRACTIONAL_Y;
	return STRETCH_FRACTIONAL_XY;
}

/* Return the description of a video configuration */
static const char* video_configuration_desc(const video_config* cfg) {
	static char buffer[256];
	double factor_x = (double)cfg->screen_visible_width / cfg->game_visible_width;
	double factor_y = (double)cfg->screen_visible_height / cfg->game_visible_height;
	unsigned perc =  cfg->game_visible_width * cfg->game_visible_height * 100 / (game_used_width * game_used_height);
	const char* stretch = "";
	switch (video_configuration_stretch(cfg)) {
		case STRETCH_INTEGER_XY : stretch = "integer"; break;
		case STRETCH_INTEGER_X_FRACTIONAL_Y : stretch = "mixed"; break;
		case STRETCH_FRACTIONAL_XY : stretch = "fract"; break;
	}
	sprintf(buffer,"%4dx%4d %2d %3d%% %4.2fx%4.2f %-7s", video_mode_size_x(cfg->mode), video_mode_size_y(cfg->mode), video_mode_bits_per_pixel(cfg->mode), perc, factor_x, factor_y, stretch);
	return buffer;
}

static void compute_result_aspect_ratio(video_config* cfg) {
	unsigned screen_aspect_ratio_x;
	unsigned screen_aspect_ratio_y;

	screen_aspect_ratio_x = video_mode_size_x(cfg->mode) * pc_aspect_ratio_y;
	screen_aspect_ratio_y = video_mode_size_y(cfg->mode) * pc_aspect_ratio_x;
	aspect_reduce(&screen_aspect_ratio_x,&screen_aspect_ratio_y);

	cfg->result_aspect_ratio_x = screen_aspect_ratio_x * cfg->screen_visible_height * cfg->game_visible_width;
	cfg->result_aspect_ratio_y = screen_aspect_ratio_y * cfg->screen_visible_width * cfg->game_visible_height;
	aspect_reduce(&cfg->result_aspect_ratio_x,&cfg->result_aspect_ratio_y);
}

/* Insert the video configuration at position gfx_cfg_mac if isn't already present */
static unsigned insert_last_if_uniq(video_config* map, unsigned mac) {
	unsigned i;
	video_config* cfg = map + mac;
	for(i=0;i<mac;++i) {
		if (map[i].mode == cfg->mode
			&& map[i].game_visible_width == cfg->game_visible_width
			&& map[i].game_visible_height == cfg->game_visible_height
			&& map[i].screen_visible_width == cfg->screen_visible_width
			&& map[i].screen_visible_height == cfg->screen_visible_height)
			return mac;
	}
	return mac + 1;
}

/*
 * Compute the resulting visible size of the game in a particular video mode
 * in:
 *   game_used_*
 *   pc_aspect_ratio_*
 *   game_aspect_ratio_*
 * out:
 *   g_* game visible size
 *   s_* screen visible size
 *   r_aspect_ratio_* resulting aspect ratio
 */
static unsigned compute_visible_raster(video_config* cfg, unsigned mac, unsigned max, video_mode* mode) {
	unsigned factor_x;
	unsigned factor_y;
	unsigned screen_aspect_ratio_x;
	unsigned screen_aspect_ratio_y;
	unsigned arcade_aspect_ratio_x;
	unsigned arcade_aspect_ratio_y;

	screen_aspect_ratio_x = video_mode_size_x(mode) * pc_aspect_ratio_y;
	screen_aspect_ratio_y = video_mode_size_y(mode) * pc_aspect_ratio_x;
	aspect_reduce(&screen_aspect_ratio_x,&screen_aspect_ratio_y);

	arcade_aspect_ratio_x = game_used_width * game_aspect_ratio_y;
	arcade_aspect_ratio_y = game_used_height * game_aspect_ratio_x;
	aspect_reduce(&arcade_aspect_ratio_x,&arcade_aspect_ratio_y);

	factor_x = screen_aspect_ratio_x * arcade_aspect_ratio_x;
	factor_y = screen_aspect_ratio_y * arcade_aspect_ratio_y;
	aspect_reduce(&factor_x,&factor_y);

	/* fractional */
	{
		if (mac == max) return mac;
		cfg[mac].mode = mode;

		/* display all the game used screen */
		cfg[mac].game_visible_width = game_used_width;
		cfg[mac].game_visible_height = game_used_height;

		/* compute screen_visible size */
		if (pc_aspect_ratio_x * arcade_aspect_ratio_y > arcade_aspect_ratio_x * pc_aspect_ratio_y) {
			/* vertical game in orizontal screen */
			cfg[mac].screen_visible_height = mode->size_y;
			cfg[mac].screen_visible_width = cfg[mac].screen_visible_height * factor_x / factor_y;
		} else {
			/* orizontal game in vertical screen */
			cfg[mac].screen_visible_width = mode->size_x;
			cfg[mac].screen_visible_height = cfg[mac].screen_visible_width * factor_y / factor_x;
		}

		compute_result_aspect_ratio(cfg + mac);
		mac = insert_last_if_uniq(cfg,mac);
	}

	/* mixed */
	{
		unsigned stretch_y_mul;
		unsigned stretch_y_div;
		unsigned stretch_x_mul;
		unsigned stretch_x_div;

		unsigned mx;
		/* approximate the request aspect with mixed integer/fractional expansion */
		/* generate all possible expansion up to fill the screen */
		for(mx=1;mx <= 1 + mode->size_x / game_used_width;++mx) {
			if (mac == max) return mac;
			cfg[mac].mode = mode;

			/* compute the x size */
			if (game_used_width * mx > mode->size_x)
				cfg[mac].game_visible_width = mode->size_x / mx;
			else
				cfg[mac].game_visible_width = game_used_width;
			cfg[mac].screen_visible_width = cfg[mac].game_visible_width * mx;

			/* compute stretch_request expansion factor in x direction */
			stretch_x_mul = cfg[mac].screen_visible_width;
			stretch_x_div = cfg[mac].game_visible_width;
			aspect_reduce(&stretch_x_mul,&stretch_x_div);

			/* compute stretch_request expansion factor in y direction for mantaining the correct aspect ratio */
			stretch_y_mul = game_aspect_ratio_x * stretch_x_mul * screen_aspect_ratio_y;
			stretch_y_div = game_aspect_ratio_y * stretch_x_div * screen_aspect_ratio_x;
			aspect_reduce(&stretch_y_mul,&stretch_y_div);

			/* compute the y size */
			cfg[mac].game_visible_height = game_used_height;
			cfg[mac].screen_visible_height = cfg[mac].game_visible_height * stretch_y_mul / stretch_y_div;

			/* overflow, correct displaying only a part of the game */
			if (cfg[mac].screen_visible_height > mode->size_y) {
				cfg[mac].screen_visible_height = mode->size_y;
				cfg[mac].game_visible_height = cfg[mac].screen_visible_height * stretch_y_div / stretch_y_mul;

				compute_result_aspect_ratio(cfg + mac);
				mac = insert_last_if_uniq(cfg,mac);

				/* generate an incorrect aspect ratio mode, but with complete game */
				if (mac == max) return mac;

				/* copy all */
				cfg[mac] = cfg[mac-1];

				/* all the game screen */
				cfg[mac].game_visible_height = game_used_height;

				compute_result_aspect_ratio(cfg + mac);
				mac = insert_last_if_uniq(cfg,mac);
			} else {
				compute_result_aspect_ratio(cfg + mac);
				mac = insert_last_if_uniq(cfg,mac);
			}
		}
	}

	/* integer */
	{
		unsigned mx,my;
		/* approximate the request aspect with integer expansion */
		/* generate all possible expansion up to fill the screen */
		for(mx=1;mx <= 1 + mode->size_x / game_used_width;++mx) {
			for(my=1;my <= 1 + mode->size_y / game_used_height;++my) {
				if (mac == max) return mac;
				cfg[mac].mode = mode;

				if (game_used_width * mx > mode->size_x)
					cfg[mac].game_visible_width = mode->size_x / mx;
				else
					cfg[mac].game_visible_width = game_used_width;
				cfg[mac].screen_visible_width = cfg[mac].game_visible_width * mx;

				if (game_used_height * my > mode->size_y)
					cfg[mac].game_visible_height = mode->size_y / my;
				else
					cfg[mac].game_visible_height = game_used_height;
				cfg[mac].screen_visible_height = cfg[mac].game_visible_height * my;

				compute_result_aspect_ratio(cfg + mac);
				mac = insert_last_if_uniq(cfg,mac);
			}
		}
	}

	/* no */
	{
		/* no stretching at all, ignore any aspect information */
		if (mac == max) return mac;
		cfg[mac].mode = mode;

		if (game_used_width > mode->size_x)
			cfg[mac].screen_visible_width = mode->size_x;
		else
			cfg[mac].screen_visible_width = game_used_width;
		cfg[mac].game_visible_width = cfg[mac].screen_visible_width;

		if (game_used_height > mode->size_y)
			cfg[mac].screen_visible_height = mode->size_y;
		else
			cfg[mac].screen_visible_height = game_used_height;
		cfg[mac].game_visible_height = cfg[mac].screen_visible_height;

		compute_result_aspect_ratio(cfg + mac);
		mac = insert_last_if_uniq(cfg,mac);
	}

	return mac;
}

static unsigned compute_visible_vector(video_config* cfg, unsigned mac, unsigned max, video_mode* mode) {
	unsigned factor_x;
	unsigned factor_y;
	unsigned screen_aspect_ratio_x;
	unsigned screen_aspect_ratio_y;
	unsigned arcade_aspect_ratio_x;
	unsigned arcade_aspect_ratio_y;

	screen_aspect_ratio_x = video_mode_size_x(mode) * pc_aspect_ratio_y;
	screen_aspect_ratio_y = video_mode_size_y(mode) * pc_aspect_ratio_x;
	aspect_reduce(&screen_aspect_ratio_x,&screen_aspect_ratio_y);

	arcade_aspect_ratio_x = game_aspect_ratio_y;
	arcade_aspect_ratio_y = game_aspect_ratio_x;
	aspect_reduce(&arcade_aspect_ratio_x,&arcade_aspect_ratio_y);

	factor_x = screen_aspect_ratio_x * arcade_aspect_ratio_x;
	factor_y = screen_aspect_ratio_y * arcade_aspect_ratio_y;
	aspect_reduce(&factor_x,&factor_y);

	if (mac == max) return mac;
	cfg[mac].mode = mode;

	/* compute screen_visible size */
	if (pc_aspect_ratio_x * arcade_aspect_ratio_y > arcade_aspect_ratio_x * pc_aspect_ratio_y) {
		/* vertical game in orizontal screen */
		cfg[mac].screen_visible_height = mode->size_y;
		cfg[mac].screen_visible_width = cfg[mac].screen_visible_height * factor_x / factor_y;
	} else {
		/* orizontal game in vertical screen */
		cfg[mac].screen_visible_width = mode->size_x;
		cfg[mac].screen_visible_height = cfg[mac].screen_visible_width * factor_y / factor_x;
	}

	/* adjust */
	cfg[mac].screen_visible_width &= ~3;
	cfg[mac].screen_visible_height &= ~3;
	cfg[mac].game_visible_width = cfg[mac].screen_visible_width;
	cfg[mac].game_visible_height = cfg[mac].screen_visible_height;

	compute_result_aspect_ratio(cfg + mac);
	mac = insert_last_if_uniq(cfg,mac);

	return mac;
}


/***************************************************************************/
/* Display mode reject */

static int reject_color(video_mode* mode) {
	switch (video_mode_bytes_per_pixel(mode)) {
		case 1 :
		case 2 :
			return 1;
	}
	return 0;
}

/* Check if a frame rate is acceptable for video sync
 * out:
 *   *rate new rate
 * return:
 *   !=0 acceptable
 *   ==0 not
 */
static int sync_accept_rate(double* rate) {
	double requested = video_fps;
	double effective = *rate;
	/* don't allow more than 8% difference between target and actual frame rate */
	while (effective > requested * 108 / 100)
		effective /= 2;
	if (effective < requested * 92 / 100)
		return 0;
	*rate = effective;
	return 1;
}

/* Check not/synced video mode */
static int reject_freq(video_mode* mode) {
	if (video_vsync_request) {
		/* vsync */
		if (video_vsync_adjust_request) {
			/* if the mode doesn't support vsync adjust reject it */
			if (!video_mode_vert_freq_adjust_avaliable(mode))
				return 0;
			if (video_monitor_mode_adjust_check(&monitor, mode, video_fps))
				return 1;
		} else {
			double rate = video_mode_vert_freq(mode);
			if (!rate)
				return 1; /* unknow frame rate */
			if (sync_accept_rate(&rate))
				return 1; /* good frame rate */
		}
	} else {
		/* not vsync */
		if (video_monitor_mode_check(&monitor, mode))
			return 1;
	}
	return 0; /* not good frame rate */
}

/* Check if a mode is acceptable or not */
static int reject(video_mode* mode) {
	/* order is not important */
	if (!reject_freq(mode))
		return 0;

	if (!reject_color(mode))
		return 0;

	return 1;
}

/***************************************************************************/
/* Display mode scoring */

/*
Video mode scoring:

Reject:
) Frequenza
) Dimensione minima per il gioco completamente visibile
) Aspetto troppo deformato

Compare:
) Aspetto del modo video
) Match numero di colori
) Frequenza simile a quella del gioco
) Tipo di modo video

*/

#if 0
/* Score scanline mode */
static int score_scanline(const video_config* cfg) {
	int have_scanline;
	switch (video_mode_driver(cfg->mode)) {
		/* small modes without the doublescan flag are considered scanline modes */
		case VIDEO_FLAGS_DRIVER_VGALINE :
			have_scanline = !cfg->mode->driver.vgaline.crtc.doublescan && cfg->mode->size_y < 480;
		break;
		case VIDEO_FLAGS_DRIVER_VBELINE :
			have_scanline = !cfg->mode->driver.vbeline.crtc.doublescan && cfg->mode->size_y < 480;
		break;
		default: /* other mode don't have scanline */
			have_scanline = 0;
	}
	if (scanlines)
		return have_scanline;
	else
		return !have_scanline;
}
#endif

/* Score the frame rate */
static long double score_continue_frequency(const video_config* cfg) {
	if (video_vsync_request && video_vsync_adjust_request) {
		/* with these option the frequency always match */
		return 0;
	} else {
		double rate = video_mode_vert_freq(cfg->mode);
		if (!rate) {
			/* unknow frame rate, use as last resource */
			return 0;
		} else {
			double effective_rate = video_fps;
			double fact;
			while (effective_rate > video_fps * 1.5)
				effective_rate /= 2;
			fact = rate/effective_rate;
			if (fact < 1.0)
				fact = 1/fact;
			return - fact*fact;
		}
	}
}

/* Score video mode too many stretched */
static long double score_continue_stretch(const video_config* cfg) {
	double fact;

	/* compute x/y distorsion */
	fact = (cfg->result_aspect_ratio_x * game_aspect_ratio_y) / (double)(cfg->result_aspect_ratio_y * game_aspect_ratio_x);
	if (fact < 1.0) fact = 1.0/fact;

	return - fact*fact;
}

/* Score game visible area */
static long double score_continue_game_visible(const video_config* cfg) {
	int width;
	int height;
	width = game_used_width;
	if (width > cfg->game_visible_width)
		width = cfg->game_visible_width;
	if (width > cfg->screen_visible_width)
		width = cfg->screen_visible_width;
	height = game_used_height;
	if (height > cfg->game_visible_height)
		height = cfg->game_visible_height;
	if (height > cfg->screen_visible_height)
		height = cfg->screen_visible_height;
	return width * height / (double)(game_used_width * game_used_height);
}

/* Score screen visible area */
static long double score_continue_screen_visible(const video_config* cfg) {
	return (cfg->screen_visible_width*cfg->screen_visible_height) / (double)(cfg->mode->size_x*cfg->mode->size_y);
}

/* Score screen size */
static long double score_continue_mode_size(const video_config* cfg) {
	return -(double)(cfg->mode->size_x*cfg->mode->size_y) / (double)(game_used_width * game_used_height);
}

/* All the continue value */
static long double score_continue(const video_config* cfg) {
	double a = gfx_cmp.game_visible * score_continue_game_visible(cfg);
	double b = gfx_cmp.mode_size * score_continue_mode_size(cfg);
	double c = gfx_cmp.screen_visible * score_continue_screen_visible(cfg);
	double d = gfx_cmp.stretch * score_continue_stretch(cfg);
	double e = gfx_cmp.frequency * score_continue_frequency(cfg);
	return a+b+c+d+e;
}

static void score_sprintf(char* buffer, const video_config* cfg) {
	sprintf(buffer,"score(%5.2f,%5.2f,%5.2f,%5.2f,%5.2f)=%6.2f",
		(double)score_continue_game_visible(cfg),
		(double)score_continue_mode_size(cfg),
		(double)score_continue_screen_visible(cfg),
		(double)score_continue_stretch(cfg),
		(double)score_continue_frequency(cfg),
		(double)score_continue(cfg)
	);
}

/* Score color depth */
static int score_color(const video_config* cfg) {
	if (game_bytes_per_pixel==2) {
		switch (video_mode_bits_per_pixel(cfg->mode)) {
			case 8 : return 0;
			case 15 : return 1;
			case 16 : return 2;
		}
	} else {
		switch (video_mode_bits_per_pixel(cfg->mode)) {
			case 8 : return 2;
			case 15 : return 0;
			case 16 : return 1;
		}
	}
	return -1;
}

/* Score vga/vesa/tweak */
static int score_display(const video_config* cfg) {
	switch (video_mode_driver(cfg->mode)) {
		case VIDEO_FLAGS_DRIVER_VBELINE :
			if (cfg->mode->driver.vbe.mode & vbeLinearBuffer)
				return 6;
			else
				return 4;
		case VIDEO_FLAGS_DRIVER_VBE :
			if (cfg->mode->driver.vbe.mode & vbeLinearBuffer)
				return 5;
			else
				return 3;
		case VIDEO_FLAGS_DRIVER_VGALINE :
			return 2;
		case VIDEO_FLAGS_DRIVER_VGA :
			return 0;
	}
	return -1;
}

/* Score size for vector game */
static int score_vector_size(const video_config* cfg) {
	return -abs(640*480 - video_mode_size_x(cfg->mode)*video_mode_size_y(cfg->mode));
}

/***************************************************************************/
/* Display mode choice */

/*
 * Compare two video mode
 * This function compare varios aspect of video mode.
 * The order of score comparison is very important.
 */
static int compare_display_mode(const video_config* a, const video_config* b) {
	long double fr;
	int r;

	r = reject(a->mode) - reject(b->mode);
	if (r) return r;

	fr = score_continue(a) - score_continue(b);
	if (fr<0) return -1;
	if (fr>0) return 1;

	r = score_color(a) - score_color(b);
	if (r) return r;

	r = score_display(a) - score_display(b);
	if (r) return r;

	logerror("video configuration compare indecision for %s, %s\n", video_configuration_name(a), video_configuration_name(b));

	return 0;
}

static int void_compare_display_mode(const void* a, const void* b) {
	return compare_display_mode((const video_config*)a,(const video_config*)b);
}

/*
 * Compare two video mode for vector game
 * This function compare varios aspect of video mode.
 * The order of score comparison is very important.
 */
static int compare_display_vector_mode(const video_config* a, const video_config* b) {
	int r;

	r = reject(a->mode) - reject(b->mode);
	if (r) return r;

	r = score_color(a) - score_color(b);
	if (r) return r;

	r = score_vector_size(a) - score_vector_size(b);
	if (r) return r;

	r = score_display(a) - score_display(b);
	if (r) return r;

	logerror("Screen video mode vector compare indecision for %s, %s\n", a->mode->name, b->mode->name);

	return 0;
}

int void_compare_display_vector_mode(const void* a, const void* b) {
	return compare_display_vector_mode((const video_config*)a,(const video_config*)b);
}

/*
 * This function sort the video configurations
 */
static void sort_video_configuration(void) {
	if (vector_game)
		qsort(gfx_cfg_map,gfx_cfg_mac,sizeof(gfx_cfg_map[0]),void_compare_display_vector_mode);
	else
		qsort(gfx_cfg_map,gfx_cfg_mac,sizeof(gfx_cfg_map[0]),void_compare_display_mode);

	{
		video_mode_container_iterator j;
		unsigned i;

		logerror("Avaliable video modes:\n");
		i = 0;
		for(video_mode_container_iterator_begin(&j,&all_gfx_modes);!video_mode_container_iterator_end(&j);video_mode_container_iterator_next(&j)) {
			video_mode* imode = video_mode_container_iterator_get(&j);
			if (reject(imode)) {
				logerror("\t[%4d] %s size %4dx%4d\n",i,video_mode_name(imode),video_mode_size_x(imode),video_mode_size_y(imode));
				++i;
			}
		}

		logerror("Score video configuration factors:\n");
		logerror("\tgame_visible %f\n",gfx_cmp.game_visible);
		logerror("\tmode_size %f\n", gfx_cmp.mode_size);
		logerror("\tscreen_visible %f\n",gfx_cmp.screen_visible);
		logerror("\tstretch %f\n", gfx_cmp.stretch);
		logerror("\tfrequency %f\n", gfx_cmp.frequency);
		logerror("Avaliable video configurations:\n");
		logerror("\tn    game      screen    mode aspect\n");
		for(i=0;i<gfx_cfg_mac;++i) {
			char score[256];
			const video_config* cfg = gfx_cfg_map + gfx_cfg_mac - 1 - i;
			score_sprintf(score,cfg);
			logerror("\t%4d %4dx%4d %4dx%4d %4dx%4d %5.2f %s %s %s\n",i,cfg->game_visible_width,cfg->game_visible_height,cfg->screen_visible_width,cfg->screen_visible_height,cfg->mode->size_x,cfg->mode->size_y,(double)cfg->result_aspect_ratio_y/cfg->result_aspect_ratio_x,score,video_configuration_name(cfg),video_mode_name(cfg->mode));
		}
	}
}

/***************************************************************************/
/* Compute interface */

/* struct contains all the information needed to inizialize */
struct osd_request {
	int width;
	int height;
	int depth;
	int fps;
	int attributes;
	int orientation;

	/* visible area */
	int min_x;
	int max_x;
	int min_y;
	int max_y;

	/* palette */
	unsigned int totalcolors;
	const unsigned char *palette;
	unsigned short *pens;
	int modifiable;
};

/*
 * Compute variables that depend on the game emulated
 */
static void compute_game_variable_raster(struct osd_request* req) {
	game_width = req->width;
	game_height = req->height;

	game_aspect_ratio_x = 1;
	game_aspect_ratio_y = 1;
	if ((req->attributes & VIDEO_PIXEL_ASPECT_RATIO_MASK) == VIDEO_PIXEL_ASPECT_RATIO_1_2)
	{
		/* These games have a double scan hardware effect */
		if (video_orientation & ORIENTATION_SWAP_XY)
			game_aspect_ratio_y *= 2;
		else
			game_aspect_ratio_x *= 2;
	}
	if ((req->attributes & VIDEO_PIXEL_ASPECT_RATIO_MASK) == VIDEO_PIXEL_ASPECT_RATIO_2_1)
	{
		/* These games have a double pixel horizontal hardware effect */
		if (video_orientation & ORIENTATION_SWAP_XY)
			game_aspect_ratio_x *= 2;
		else
			game_aspect_ratio_y *= 2;
	}

	game_used_posx = req->min_x;
	game_used_posy = req->min_y;
	game_used_width = req->max_x - req->min_x + 1;
	game_used_height = req->max_y - req->min_y + 1;

	/* convert all game_* variables to match the blit_orientation */
	if (blit_orientation & ORIENTATION_SWAP_XY) {
		int temp;
		temp = game_used_width;
		game_used_width = game_used_height;
		game_used_height = temp;
		temp = game_width;
		game_width = game_height;
		game_height = temp;
		temp = game_used_posx;
		game_used_posx = game_used_posy;
		game_used_posy = temp;
		temp = game_aspect_ratio_x;
		game_aspect_ratio_x = game_aspect_ratio_y;
		game_aspect_ratio_y = temp;
	}

	if (blit_orientation & ORIENTATION_FLIP_X) {
		game_used_posx = game_width - game_used_posx - game_used_width;
	}

	if (blit_orientation & ORIENTATION_FLIP_Y) {
		game_used_posy = game_height - game_used_posy - game_used_height;
	}

	/* round to a multiple of 4 to avoid missing pixels on the right side */
	game_used_width = (game_used_width + 3) & ~3;
}

/*
 * Compute variables that depend on the game emulated
 */
static void compute_game_variable_vector(struct osd_request* req) {
	game_width = 0;
	game_height = 0;
	if (video_orientation & ORIENTATION_SWAP_XY) {
		game_aspect_ratio_x = req->width;
		game_aspect_ratio_y = req->height;
	} else {
		game_aspect_ratio_x = req->height;
		game_aspect_ratio_y = req->width;
	}
	game_used_posx = 0;
	game_used_posy = 0;
	game_used_width = 0;
	game_used_height = 0;
}

/*
 * Compute variables that depend on the game emulated
 */
static void compute_game_variable(struct osd_request* req) {
	game_bits_per_pixel = req->depth;
	game_bytes_per_pixel = (game_bits_per_pixel + 7) / 8;

	/* look if this is a vector game */
	if (video_attributes & VIDEO_TYPE_VECTOR)
		vector_game = 1;
	else
		vector_game = 0;
}

/*
 * Compute variables that don't depend from game or video mode
 */
static int compute_state_variable(struct osd_request* req) {
	video_mode_container_iterator i;

	video_fps = req->fps;
	video_attributes = req->attributes;
	video_orientation = req->orientation;

	if (frameskip < 0) frameskip = 0;
	if (frameskip >= FRAMESKIP_LEVELS) frameskip = FRAMESKIP_LEVELS-1;

	gone_to_gfx_mode = 0;

	/* load monitor config */
	if (video_monitor_load(&monitor)!=0) {
		fprintf(stderr,"Error loading the monitor options\n%s\n", video_error_description_get());
		fprintf(stderr,"Plase read the file `install.txt' and `mv.txt'\n");
		return 0;
	}

	/* load graphics modes */
	video_mode_container_init(&all_gfx_modes);

	if (video_mode_container_load(&all_gfx_modes)!=0) {
		fprintf(stderr,
"MAME have detected one unsupported or invalid video mode in your 'modeline'\n"
"configuration.\n"
"If you have manually edited these line probably you have made a mistake.\n"
"Otherwise probably is changed your VBE (VESA) support and the old video mode\n"
"numbering became inconsistent, in this case manually remove all the VBE\n"
"modelines from the mame.cfg file.\n\n");
		fprintf(stderr,"%s\n",video_error_description_get());
		return 0;
	}

	gfx_cfg = 0;

	video_mode_container_iterator_begin(&i,&all_gfx_modes);
	if (video_mode_container_iterator_end(&i)) {
		fprintf(stderr,"No video modes present in your configuration file\n");
		fprintf(stderr,"Plase read the file `install.txt' and `mv.txt'\n");
		return 0;
	}

	brightness_paused_adjust = 1.0;
	bright_dirty = 1;

	/* inizialize the turbo frame counter */
	turbo_frame_counter = 0;

	return 1;
}

/*
	Create the set of avaliable video configurations
*/
static int compute_video_configuration(void) {
	video_mode_container_iterator j;

	/* create all possible graphics configurations */
	gfx_cfg_mac = 0;

	for(video_mode_container_iterator_begin(&j,&all_gfx_modes);!video_mode_container_iterator_end(&j);video_mode_container_iterator_next(&j)) {
		video_mode* imode = video_mode_container_iterator_get(&j);
		if (vector_game)
			gfx_cfg_mac = compute_visible_vector(gfx_cfg_map, gfx_cfg_mac, GFX_CFG_MAX, imode);
		else
			gfx_cfg_mac = compute_visible_raster(gfx_cfg_map, gfx_cfg_mac, GFX_CFG_MAX, imode);

	}

	if (gfx_cfg_mac == 0) {
		fprintf(stderr,"No video modes avaliable\n");
		return 0;
	}

	return 1;
}

/* Check if a video configuration is acceptable */
static int check_video_configuration(const video_config* gfx) {
	assert( gfx && gfx != gfx_cfg_map + gfx_cfg_mac && gfx->mode);
 	return reject(gfx->mode);
}

/* Check if a video configuration is acceptable with the default request */
static int check_auto_video_configuration(const video_config* gfx) {
	if (vector_game)
		return 1;
	if (stretch_request == STRETCH_FRACTIONAL_XY) {
		/* all ok */
	} else if (stretch_request == STRETCH_INTEGER_X_FRACTIONAL_Y) {
		if (gfx->screen_visible_width % gfx->game_visible_width != 0)
			return 0;
	} else if (stretch_request == STRETCH_INTEGER_XY) {
		if (gfx->screen_visible_width % gfx->game_visible_width != 0
			|| gfx->screen_visible_height % gfx->game_visible_height != 0)
			return 0;
	} else if (stretch_request == STRETCH_NONE) {
		if (gfx->screen_visible_width != gfx->game_visible_width
			|| gfx->screen_visible_height != gfx->game_visible_height)
			return 0;
	}
	return 1;
}

static video_config* choice_best_video_configuration_base(void) {
	video_config* cfg = 0;

	if (stricmp(resolution_request,"auto")!=0) {
		int i;
		for(i=gfx_cfg_mac-1;i>=0;--i) {
			if ( (stricmp(resolution_request,video_mode_name(gfx_cfg_map[i].mode)) == 0
				|| stricmp(resolution_request,video_configuration_name(&gfx_cfg_map[i])) == 0)
				&& check_video_configuration(&gfx_cfg_map[i])) {
				cfg = gfx_cfg_map + i;
				break;
			}
		}
	} else {
		int i;
		for(i=gfx_cfg_mac-1;i>=0;--i) {
			if (check_video_configuration(&gfx_cfg_map[i])
				&& check_auto_video_configuration(&gfx_cfg_map[i])) {
				cfg = gfx_cfg_map + i;
				break;
			}
		}
	}

	return cfg;
}

static video_config* choice_best_video_configuration(void) {
	video_config* cfg;

	logerror("Select the best video configuration:\n");

	cfg = choice_best_video_configuration_base();

	if (!cfg && video_vsync_request) {
		video_vsync_request = 0;
		logerror("warning: No video configurations avaliable, retrying with the 'vsync' option disabled\n");

		cfg = choice_best_video_configuration_base();
	}

	if (!cfg && stricmp(resolution_request,"auto")!=0) {
		strcpy(resolution_request,"auto");
		logerror("warning: No video configurations avaliable, retrying with the 'video' option disabled\n");

		cfg = choice_best_video_configuration_base();
	}

	if (!cfg) {
		fprintf(stderr,"No video modes avaliable for this game\n");
		return 0;
	}

	logerror("\tname %s\n",video_configuration_name(cfg));

	return cfg;
}

/*
	Adjust game variables for the chosen video configuration
*/
static void adjust_video_configuration(void) {
	/* adjust for vector game */
	if (vector_game) {
		game_used_width = game_width = gfx_cfg->game_visible_width;
		game_used_height = game_height = gfx_cfg->game_visible_height;

		/* signal to the core the size of the vector game */
		if (video_orientation & ORIENTATION_SWAP_XY) {
			options.vector_width = game_height;
			options.vector_height = game_width;
		} else {
			options.vector_width = game_width;
			options.vector_height = game_height;
		}
	}

	logerror("Compute game configuration:\n");
	logerror("\tmachine_orientation %s%s%s\n", video_orientation & ORIENTATION_SWAP_XY ? "ORIENTATION_SWAP_XY " : "", video_orientation & ORIENTATION_FLIP_X ? "ORIENTATION_FLIP_X " : "", video_orientation & ORIENTATION_FLIP_Y ? "ORIENTATION_FLIP_Y " : "");
	logerror("\tbits_per_pixel %d\n", game_bits_per_pixel);
	logerror("\tsize %dx%d\n", game_width, game_height);
	logerror("\tsize_used %dx%d\n", game_used_width, game_used_height);
	logerror("\taspect_ratio %d/%d\n", game_aspect_ratio_x, game_aspect_ratio_y);
}

/*
 * Compute all static variable which directly depend on the video mode
 */
static void compute_video_variable(void)
{
	logerror("Compute video configuration:\n");

	assert( gfx_cfg );

	assert( gfx_cfg->game_visible_width <= game_used_width && gfx_cfg->game_visible_height <= game_used_height );

	/* set stretching parameter */
	if (gfx_cfg->screen_visible_width == gfx_cfg->game_visible_width && gfx_cfg->screen_visible_height == gfx_cfg->game_visible_height) {
		screen_stretch = SCREEN_STRETCH_NONE;
		screen_stretch_factorx = 1;
		screen_stretch_factory = 1;
	} else if (gfx_cfg->screen_visible_width % gfx_cfg->game_visible_width == 0 && gfx_cfg->screen_visible_height % gfx_cfg->game_visible_height == 0) {
		screen_stretch = SCREEN_STRETCH_INTEGER;
		screen_stretch_factorx = gfx_cfg->screen_visible_width / gfx_cfg->game_visible_width;
		screen_stretch_factory = gfx_cfg->screen_visible_height / gfx_cfg->game_visible_height;
	} else {
		screen_stretch = SCREEN_STRETCH_FRACTIONAL;
		screen_stretch_factorx = 0; /* not used */
		screen_stretch_factory = 0;
	}

	logerror("\tname %s\n", video_configuration_name(gfx_cfg));
	logerror("\tblit_orientation %s%s%s\n", blit_orientation & ORIENTATION_SWAP_XY ? "ORIENTATION_SWAP_XY " : "", blit_orientation & ORIENTATION_FLIP_X ? "ORIENTATION_FLIP_X " : "", blit_orientation & ORIENTATION_FLIP_Y ? "ORIENTATION_FLIP_Y " : "");
	logerror("\tvisible_size %dx%d\n", gfx_cfg->game_visible_width, gfx_cfg->game_visible_height);
	logerror("\taspect_ratio %d/%d\n", gfx_cfg->result_aspect_ratio_x, gfx_cfg->result_aspect_ratio_y);
	logerror("\tstretch x*%.2f y*%.2f\n", (double)gfx_cfg->screen_visible_width / gfx_cfg->game_visible_width, (double)gfx_cfg->screen_visible_height / gfx_cfg->game_visible_height);
	logerror("\tscanlines %s\n", scanlines ? "yes" : "no" );

	/* inizialize the dirty system */
	{
		/* one level for any video buffer */
		unsigned level = video_update_page_max();
		dirty_init( level, (gfx_cfg->game_visible_width + DIRTY_DELTA_X - 1) / DIRTY_DELTA_X, (gfx_cfg->game_visible_height + DIRTY_DELTA_Y - 1) / DIRTY_DELTA_Y );
	}

	/* screen position */
	blit_x = (video_size_x() - gfx_cfg->screen_visible_width) / 2;
	blit_y = (video_size_y() - gfx_cfg->screen_visible_height) / 2;

	if (video_unchained_active()) {
		unsigned pixel = ALIGN_UNCHAINED / video_bytes_per_pixel();
		blit_x = (blit_x + pixel-1) & ~(pixel-1);
	} else {
		unsigned pixel = ALIGN / video_bytes_per_pixel();
		blit_x = (blit_x + pixel-1) & ~(pixel-1);
	}

	/* source increment */
	blit_dw = (char*)Machine->scrbitmap->line[1] - (char*)Machine->scrbitmap->line[0];
	blit_dp = game_bytes_per_pixel;

	if (blit_orientation & ORIENTATION_SWAP_XY) {
		int temp;
		temp = blit_dw;
		blit_dw = blit_dp;
		blit_dp = temp;
	}

	blit_offset = 0;

	if (blit_orientation & ORIENTATION_FLIP_Y) {
		blit_offset += (game_height - 1) * blit_dw;
		blit_dw = -blit_dw;
	}

	if (blit_orientation & ORIENTATION_FLIP_X) {
		blit_offset += (game_width - 1) * blit_dp;
		blit_dp = -blit_dp;
	}

	/* compute the source x position aligment */
	if (blit_orientation & ORIENTATION_SWAP_XY) {
		/* alignement not requiried */
		game_visible_posx_increment = 1;
	} else {
		assert(ALIGN % abs(blit_dp) == 0);
		game_visible_posx_increment = ALIGN / abs(blit_dp);
		if (!game_visible_posx_increment) {
			game_visible_posx_increment = 1;
		}
	}

	/* pipeline */
	assert( blit_pipeline_flag == 0 );
}

static void free_video_variable(void)
{
	dirty_done();

	if (blit_pipeline_flag) {
		video_pipeline_done(&blit_pipeline);
		blit_pipeline_flag = 0;
	}
}

/*
 * Activate the chosen video configuration
 */
static int activate_video_configuration(void)
{
	unsigned i;
	assert( gfx_cfg );

	if (video_vsync_request
		&& video_vsync_adjust_request
		&& video_mode_vert_freq_adjust_avaliable(gfx_cfg->mode)) {
		/* try to program the mode to the exact frame rate */
		logerror("VSync adjust:\n");
		logerror("\toriginal frame rate %.2f\n",video_mode_vert_freq(gfx_cfg->mode));
		video_mode_vert_freq_adjust(gfx_cfg->mode,video_fps);
		logerror("\tnew frame rate %.2f\n",video_mode_vert_freq(gfx_cfg->mode));
	}

	/* final paranoic check */
	if (!video_monitor_mode_check(&monitor, gfx_cfg->mode)) {
		fprintf(stderr,"Unexpected error. The selected video mode isn't in the frequencies range of your monitor.\n");
		return 0;
	}

	/* set the video mode */
	if (video_mode_set(gfx_cfg->mode)!=0) {
		logerror("Error activating video mode %s\n", video_mode_name(gfx_cfg->mode) );
		return 0;
	}

	/* inizialize the stretch_request system */
	video_stretch_init();

	/* inizialize the update system */
	video_update_init( use_triplebuf_request!=0 ? 3 : 1 );

	/* clear the all the screens */
	/* (can't be cleared the entire video memory for some strange graphics cards like neomagic) */
	for(i=0;i<video_update_page_max();++i) {
		video_update_start();
		video_clear(video_update_x(),video_update_y(),video_size_x(), video_size_y(),0);
		video_update_stop(0);
	}

	logerror("Compute video mode:\n");
	logerror("\tname %s\n", video_name() );
	logerror("\tsize %dx%d\n", video_size_x(), video_size_y());
	logerror("\tbits_per_pixel %d\n", video_bits_per_pixel());
	logerror("\tbytes_per_scanline %d\n", video_bytes_per_scanline());
	logerror("\tpages %d\n", video_update_page_max());

	gone_to_gfx_mode = 1;

	return 1;
}

static void deactivate_video_configuration(int reset_video_mode) {
	if (gone_to_gfx_mode != 0)
	{
		gone_to_gfx_mode = 0;

		video_update_done();
		video_stretch_done();
		if (reset_video_mode)
			video_mode_reset();
	}
}

/***************************************************************************/
/* sync */

static int compute_sync(void) {
	warming_up = 1;

	logerror("Compute vsync:\n");
	logerror("\tvsync %s\n", video_vsync_request ? "yes" : "no");

	vsync_frame_rate = video_fps;

	if (video_vsync_request)
	{
		/* measure frame rate */
		double test_rate = video_vert_freq();

		logerror("\tVideo mode frequency %3.2fHz\n", test_rate);
		logerror("\tArcade original frequency %3.2fHz\n", (double)video_fps);

		if (!test_rate) {
			double rate = video_measure_vert_freq();

			logerror("\tVideo mode measured frequency %3.2fHz\n",rate);

			if (rate < 10 || rate > 200) {
				osd_close_display();
				fprintf(stderr,"Error measuring the video vertical frequency (frame rate).\n");
				fprintf(stderr,"\tVideo mode measured frequency %3.2fHz\n",rate);
				return 0;
			}

			test_rate = rate;
		}

		if (!sync_accept_rate(&test_rate))
		{
			osd_close_display();
			fprintf(stderr,"Error in the video mode vertical frequency (frame rate)\n");
			fprintf(stderr,"If you activate the option \"vsync\" a video mode that match the\n");
			fprintf(stderr,"original arcade vertical frequency is required.\n");
			fprintf(stderr,"Try to use the waitvsync option instead\n");
			fprintf(stderr,"\tVideo mode frequency %3.2fHz\n",test_rate);
			fprintf(stderr,"\tArcade original frequency %3.2fHz\n", (double)video_fps);
			return 0;
		}

		/* TODO this issue need testing and comparing with VSyncMAME */
		/* assume a small error in the measure */
		/* otherwise a lost of sync may happen causing a big slowdown */
		vsync_frame_rate = test_rate * 0.99;

		logerror("\tAdjusted frequency %3.2fHz\n",vsync_frame_rate);

		if (Machine->sample_rate)
		{
			Machine->sample_rate = Machine->sample_rate * video_fps / vsync_frame_rate;
			logerror("\tAdjusted sample rate %d\n", (int)Machine->sample_rate);
		}
	}

	return 1;
}

/***************************************************************************/
/* skip */

static void save_skip(const char* section) {
	if (skipcolumns_request>=0)
		set_config_int((char*)section, "skipcolumns", skipcolumns_request);
	else
		set_config_string((char*)section, "skipcolumns", "auto");
	if (skiplines_request>=0)
		set_config_int((char*)section, "skiplines", skiplines_request);
	else
		set_config_string((char*)section, "skiplines", "auto");
}

static void compute_skip(void) {
	int posx,posy,sizex,sizey;

	logerror("Compute skip:\n");

	/* center if skips are negative */
	if (skipcolumns_request<0)
		game_visible_posx = (game_used_width - gfx_cfg->game_visible_width) / 2;
	else
		game_visible_posx = skipcolumns_request;
	if (skiplines_request<0)
		game_visible_posy = (game_used_height - gfx_cfg->game_visible_height) / 2;
	else
		game_visible_posy = skiplines_request;

	/* Failsafe against silly parameters */
	if (game_visible_posx < 0) {
		game_visible_posx = 0;
		logerror("warning: adjusted game_visible_posx\n");
	}
	if (game_visible_posy < 0) {
		game_visible_posy = 0;
		logerror("warning: adjusted game_visible_posy\n");
	}
	if (game_visible_posx + gfx_cfg->game_visible_width > game_used_width) {
		game_visible_posx = game_used_width - gfx_cfg->game_visible_width;
		logerror("warning: adjusted game_visible_posx\n");
	}
	if (game_visible_posy + gfx_cfg->game_visible_height > game_used_height) {
		game_visible_posy = game_used_height - gfx_cfg->game_visible_height;
		logerror("warning: adjusted game_visible_posy\n");
	}

	/* adjust the source position */
	game_visible_posx = game_visible_posx -	game_visible_posx % game_visible_posx_increment;

	logerror("\tskip %dx%d\n", game_visible_posx, game_visible_posy);

	/* the skipcolumns from mame.cfg/cmdline is relative to the visible area */
	game_visible_posx += game_used_posx;
	game_visible_posy += game_used_posy;

	posx = game_visible_posx;
	posy = game_visible_posy;
	sizex = gfx_cfg->game_visible_width;
	sizey = gfx_cfg->game_visible_height;

	if (blit_orientation & ORIENTATION_FLIP_X) {
		posx = game_width - sizex - posx;
	}

	if (blit_orientation & ORIENTATION_FLIP_Y) {
		posy = game_height - sizey - posy;
	}

	if (blit_orientation & ORIENTATION_SWAP_XY) {
		int temp = posx;
		posx = posy;
		posy = temp;
		temp = sizex;
		sizex = sizey;
		sizey = temp;
	}

	logerror("\tui visible area %d,%d %d,%d\n",posx,posy,posx+sizex-1,posy+sizey-1);

	set_ui_visarea(posx,posy,posx+sizex-1,posy+sizey-1);
}

void change_skip(void) {
	compute_skip();
	if (dirty_effective)
		dirty_set_all();
}

/***************************************************************************/
/* stretch */

static void save_stretch(const char* section) {
	const char* s = 0;
	switch (stretch_request) {
		case STRETCH_NONE : s = "no"; break;
		case STRETCH_INTEGER_XY : s = "integer"; break;
		case STRETCH_INTEGER_X_FRACTIONAL_Y : s = "mixed"; break;
		case STRETCH_FRACTIONAL_XY : s = "fractional"; break;
	}
	set_config_string((char*)section, (char*)"resize", (char*)s);
}

static void clear_stretch(const char* section) {
	set_config_string((char*)section, (char*)"resize", 0);
}

/***************************************************************************/
/* wait_vsync */

static void save_wait(const char* section) {
	const char* s = 0;
	switch (wait_vsync_request) {
		case 0 : s = "no"; break;
		case 1 : s = "yes"; break;
	}
	set_config_string((char*)section, (char*)"waitvsync", (char*)s);
}

/* Update teh wait_vsync variable */
static void compute_wait(void) {
	wait_vsync_effective = wait_vsync_request;

	logerror("Compute waitvsync:\n");

	if (wait_vsync_effective && video_vsync_request)
	{
		wait_vsync_effective = 0;
		logerror("warning: waitvsync disabled because vsync is already active\n");
	}

	logerror("\twaitvsync %s\n", wait_vsync_effective ? "yes" : "no" );
}

/***************************************************************************/
/* dirty */

/*
static void save_dirty(const char* section) {
	const char* s = 0;
	switch (dirty_request) {
		case -1 : s = "auto"; break;
		case 0 : s = "no"; break;
		case 1 : s = "yes"; break;
	}
	set_config_string((char*)section, (char*)"dirty", (char*)s);
}
*/

/* Update the dirty_effective variable */
static void compute_dirty(void) {
	logerror("Compute dirty:\n");

	/* Is the game using a dirty system? */
	if ((video_attributes & VIDEO_SUPPORTS_DIRTY)!=0 && dirty_request!=0)
		dirty_effective = 1;
	else
		dirty_effective = 0;

	if (dirty_effective && screen_stretch==SCREEN_STRETCH_FRACTIONAL) {
		logerror("warning: dirty disabled because a not integer stretch_request factor is used\n");
		dirty_effective = 0;
	}

	if (dirty_effective && blit_combine_effective != COMBINE_NONE) {
		logerror("warning: dirty disabled because bliteffect!=no\n");
		dirty_effective = 0;
	}

	logerror("\tdirty %s\n", dirty_effective ? "yes" : "no" );
}

/***************************************************************************/
/* combine */

static const char* GAME_BLIT_COMBINE_MAX[] = {
#include "blitmax.h"
0
};

/* defined in config.c */
int is_game_in_list(const char* list[], const struct GameDriver* pgame);

static void save_combine(const char* section) {
	const char* s = 0;
	switch (blit_combine_request) {
		case COMBINE_AUTO : s = "auto"; break;
		case COMBINE_NONE : s = "no"; break;
		case COMBINE_MEAN : s = "mean"; break;
		case COMBINE_MAX : s = "max"; break;
		case COMBINE_FILTER : s = "filter"; break;
		case COMBINE_FILTERX : s = "filterx"; break;
		case COMBINE_FILTERY : s = "filtery"; break;
	}
	set_config_string((char*)section, (char*)"bliteffect", (char*)s);
}

/* Update the blit_combine_effective variable */
static void compute_combine(void) {
	assert( palette_total );

	logerror("Compute effect:\n");

	blit_combine_effective = blit_combine_request;

	if (blit_combine_effective == COMBINE_AUTO) {
		if (is_game_in_list(GAME_BLIT_COMBINE_MAX,Machine->gamedrv)) {
			logerror("\tbliteffect (auto) max\n");
			blit_combine_effective = COMBINE_MAX;
		} else {
			logerror("\tbliteffect (auto) mean\n");
			blit_combine_effective = COMBINE_MEAN;
		}
	}

/*
	if (blit_combine_effective == COMBINE_MAX && palette_rgb) {
		blit_combine_effective = COMBINE_NONE;
		logerror("warning: bliteffect=max forced to =no because no soft/hard palette is used\n");
	}
*/

	if ((blit_combine_effective == COMBINE_MEAN || blit_combine_effective == COMBINE_FILTER || blit_combine_effective == COMBINE_FILTERX || blit_combine_effective == COMBINE_FILTERY) &&
		video_index() != VIDEO_FLAGS_INDEX_RGB) {
		blit_combine_effective = COMBINE_NONE;
		logerror("warning: bliteffect=mean|filter forced to =no because the video mode is not hardware rgb\n");
	}

	if (blit_combine_effective == COMBINE_MAX && palette_modifiable &&
		video_index() != VIDEO_FLAGS_INDEX_RGB) {
		blit_combine_effective = COMBINE_NONE;
		logerror("warning: bliteffect=max forced to =no because the game palette is modifiable and the mode active isn't rgb\n");
	}

	if ((blit_combine_effective == COMBINE_MAX || blit_combine_effective == COMBINE_MEAN)
		&& screen_stretch == SCREEN_STRETCH_NONE) {
		blit_combine_effective = COMBINE_NONE;
		logerror("warning: bliteffect=mean|max forced to =no because no stretching activated\n");
	}

	switch (blit_combine_effective) {
		case COMBINE_MAX : logerror("\tbliteffect max\n"); break;
		case COMBINE_MEAN : logerror("\tbliteffect mean\n"); break;
		case COMBINE_FILTER : logerror("\tbliteffect filter\n"); break;
		case COMBINE_FILTERX : logerror("\tbliteffect filterx\n"); break;
		case COMBINE_FILTERY : logerror("\tbliteffect filtery\n"); break;
		case COMBINE_NONE :  logerror("\tbliteffect no\n"); break;
	}

	/* destroy the pipeline if already created */
	if (blit_pipeline_flag) {
		video_pipeline_done(&blit_pipeline);
		blit_pipeline_flag = 0;
	}
}

/***************************************************************************/
/* resolution */

/* Update all changeable variable */
static void compute_changeable(void) {
	compute_wait();
	compute_combine();
	compute_dirty();
	compute_skip();
}

static void save_resolution(const char* section) {
	/* update the video option config file */
	set_config_string((char*)section, "video", resolution_request);
}

static void compute_resolution(video_config* cfg) {
	int i;

	free_video_variable();

	if (cfg->mode != gfx_cfg->mode) {
		deactivate_video_configuration(0);
		gfx_cfg = cfg;
		adjust_video_configuration();
		activate_video_configuration(); /* TODO check error */
	} else {
		gfx_cfg = cfg;
		adjust_video_configuration();

		/* clear the all the screens */
		for(i=0;i<video_update_page_max();++i) {
			video_update_start();
			video_clear(video_update_x(),video_update_y(),video_size_x(), video_size_y(),0);
			video_update_stop(0);
		}
	}

	/* don't call compute_sync() */
	compute_video_variable();

	/* Recompute the palette */
	if (palette_rgb) {
		if (video_index() == VIDEO_FLAGS_INDEX_PACKED)
			video_index_packed_to_rgb(0);
	} else {
		free(palette_lookup);
		palette_lookup = 0;

		/* if too many colors for the hardware palette, use the rgb direct color mode */
		if (video_index() == VIDEO_FLAGS_INDEX_PACKED && palette_total > (1U << video_bits_per_pixel()))
			video_index_packed_to_rgb(0);

		/* if not hardware palette support, create the software palette */
		if (video_index() == VIDEO_FLAGS_INDEX_RGB) {
			palette_lookup = malloc(palette_total * sizeof(unsigned));
				assert(palette_lookup);
		}

		/* mark the palette dirty */
		palette_dirty = 1;
		for (i = 0;i < palette_total;i++)
			palette_dirty_color[i] = 1;
	}

	compute_changeable();
}

/**************************************************************************/
/* state */

void change_state(void) {
	video_config* cfg = choice_best_video_configuration();
	if (cfg != gfx_cfg)
		compute_resolution(cfg);
	else
		compute_changeable();
	if (dirty_effective)
		dirty_set_all();
}

void save_state(const char* section) {
	save_resolution(section);
	if (!vector_game) {
		if (strcmp(resolution_request,"auto")!=0)
			clear_stretch(section);
		else
			save_stretch(section);
		save_skip(section);
		save_combine(section);
	}
	save_wait(section);

	usrintf_showmessage("Video options saved in [%s]",section);
}

static int color_compare(const void* pA, const void* pB) {
	unsigned A = *(const unsigned*)pA;
	unsigned B = *(const unsigned*)pB;
	int Av = (unsigned)current_palette[A].red + (unsigned)current_palette[A].green + (unsigned)current_palette[A].blue;
	int Bv = (unsigned)current_palette[B].red + (unsigned)current_palette[B].green + (unsigned)current_palette[B].blue;

	if (current_palette[A].alpha < current_palette[B].alpha)
		return -1;
	if (current_palette[A].alpha > current_palette[B].alpha)
		return 1;

	if (Av < Bv)
		return -1;
	if (Av > Bv)
		return 1;

	return 0;
}

void palette_sort(unsigned short *pens, unsigned pens_total) {
	unsigned i;

	video_color* palette = malloc(sizeof(video_color) * palette_total);
	unsigned* tbl_index = malloc(sizeof(unsigned) * palette_total);
	unsigned* tbl_rindex = malloc(sizeof(unsigned) * palette_total);

	assert( palette && tbl_index );

	for(i=0;i<palette_total;++i) {
		tbl_index[i] = i;
		palette[i] = current_palette[i];
	}

	qsort(tbl_index,palette_total,sizeof(tbl_index[0]),color_compare);

	for(i=0;i<palette_total;++i) {
		current_palette[i] = palette[tbl_index[i]];
		tbl_rindex[tbl_index[i]] = i;
	}

	for(i=0;i<pens_total;++i) {
		pens[i] = tbl_rindex[pens[i]];
	}

	free(palette);
	free(tbl_index);
	free(tbl_rindex);
}

int compute_palette(struct osd_request* req)
{
	int i;

	palette_modifiable = req->modifiable;
	palette_total = 0;
	palette_dirty = 0;
	palette_rgb = 0;

	assert( !palette_dirty_color && !current_palette && !palette_lookup );

	logerror("Compute palette:\n");
	logerror("\tcolors %d\n",req->totalcolors);
	logerror("\tmodifiable %s\n", req->modifiable ? "yes" : "no");

	if (video_index() == VIDEO_FLAGS_INDEX_RGB
		&& video_bytes_per_pixel() == game_bytes_per_pixel
		&& !palette_modifiable)
	{
		/* Don't use a palette, pens contain direct rgb colors */
		int r,g,b;
		video_rgb rgb;

		logerror("\tmode hardware rgb\n");

		palette_rgb = 1;

		for (i = 0;i < req->totalcolors;i++)
		{
			r = 255 * brightness * pow(req->palette[3*i+0] / 255.0, 1 / osd_gamma_correction) / 100;
			if (r>255) r = 255;
			g = 255 * brightness * pow(req->palette[3*i+1] / 255.0, 1 / osd_gamma_correction) / 100;
			if (g>255) g = 255;
			b = 255 * brightness * pow(req->palette[3*i+2] / 255.0, 1 / osd_gamma_correction) / 100;
			if (b>255) b = 255;
			video_rgb_make(&rgb,r,g,b);
			*req->pens++ = rgb;
		}

		video_rgb_make(&rgb,0x00,0x00,0x00);
		Machine->uifont->colortable[0] = rgb;
		video_rgb_make(&rgb,0xff,0xff,0xff);
		Machine->uifont->colortable[1] = rgb;
		video_rgb_make(&rgb,0xff,0xff,0xff);
		Machine->uifont->colortable[2] = rgb;
		video_rgb_make(&rgb,0x00,0x00,0x00);
		Machine->uifont->colortable[3] = rgb;
	}
	else
	{
		/* Use a palette, hardware or software, pens contain the palette index */
		int usercolors; /* colors for the user interface */

		palette_rgb = 0;

		/* add space for user interface color */
		if ((video_index() == VIDEO_FLAGS_INDEX_PACKED && video_bytes_per_pixel()==1)
			|| game_bytes_per_pixel==1) {
			/* if games or video is limited at 8 bit don't overflow */
			if (req->totalcolors==256) {
				usercolors = 0;
			} else if (req->totalcolors==255) {
				usercolors = 1;
			} else {
				usercolors = 2;
			}
		} else {
			usercolors = 2;
		}

		palette_total = req->totalcolors + usercolors;
		palette_dirty_color = (int*)malloc(palette_total * sizeof(int));
		current_palette = (video_color*)malloc(palette_total * sizeof(video_color));

		assert( palette_dirty_color && current_palette );

		for (i = 0;i < palette_total;i++)
			palette_dirty_color[i] = 1;
		palette_dirty = 1;
		for (i = 0;i < palette_total;i++) {
			current_palette[i].red = 0;
			current_palette[i].green = 0;
			current_palette[i].blue = 0;
		}

		/* if too many colors for the hardware palette, use the rgb direct color mode */
		if (video_index() == VIDEO_FLAGS_INDEX_PACKED && palette_total > (1U << (video_bits_per_pixel())))
			video_index_packed_to_rgb(0);

		/* if not hardware palette support, create the software palette */
		if (video_index() == VIDEO_FLAGS_INDEX_RGB) {
			palette_lookup = malloc(palette_total * sizeof(unsigned));
			assert(palette_lookup);
		}

		/* add user interface colors */
		if (usercolors > 0) {
			current_palette[0].red = 0x0;
			current_palette[0].green = 0x0;
			current_palette[0].blue = 0x0;
			current_palette[0].alpha = 0; /* first color */
		}
		if (usercolors > 1) {
			current_palette[1].red = 0xff;
			current_palette[1].green = 0xff;
			current_palette[1].blue = 0xff;
			current_palette[1].alpha = 2; /* last color */
		}

		for (i = 0;i < req->totalcolors;i++)
			req->pens[i] = usercolors + i;

		for (i = 0;i < req->totalcolors;i++)
		{
			current_palette[req->pens[i]].red = req->palette[3*i+0];
			current_palette[req->pens[i]].green = req->palette[3*i+1];
			current_palette[req->pens[i]].blue = req->palette[3*i+2];
			current_palette[req->pens[i]].alpha = 1; /* middle color */
		}

		/* sort pens from black to white, used for max blit */
		palette_sort(req->pens, req->totalcolors);

		/* UI color */
		Machine->uifont->colortable[0] = 0;
		Machine->uifont->colortable[1] = palette_total - 1;
		Machine->uifont->colortable[2] = palette_total - 1;
		Machine->uifont->colortable[3] = 0;

		if (video_index() == VIDEO_FLAGS_INDEX_RGB)
			logerror("\tmode hardware rgb, software palette\n");
		else
			logerror("\tmode hardware palette\n");

	}

	return 1;
}

static int compute_mode(void) {
	if (!compute_video_configuration())
		return 0;

	sort_video_configuration();

	gfx_cfg = choice_best_video_configuration();
	if (!gfx_cfg)
		return 0;

	adjust_video_configuration();

	return 1;
}

static void free_all(void)
{
	free_video_variable();
	deactivate_video_configuration(1);

	if (frames_displayed > FRAMES_TO_SKIP)
		printf("Average FPS: %f\n",(double)TICKS_PER_SEC/(end_time-start_time)*(frames_displayed-FRAMES_TO_SKIP));

	free(palette_dirty_color);
	palette_dirty_color = 0;
	free(current_palette);
	current_palette = 0;
	free(palette_lookup);
	palette_lookup = 0;
}

/***************************************************************************/
/* OSD interface */

static struct osd_request OSD_REQ;

int osd_init_display(int width,int height,int depth,int fps,int attributes,int orientation) {
	OSD_REQ.width = width;
	OSD_REQ.height = height;
	OSD_REQ.depth = depth;
	OSD_REQ.fps = fps;
	OSD_REQ.attributes = attributes;
	OSD_REQ.orientation = orientation;

	if (!compute_state_variable(&OSD_REQ))
		return -1;

	compute_game_variable(&OSD_REQ);

	/* Compute early the video mode for vector game */
	/* In this way the core can size the vector bitmap exactly as the video mode chosen */
	if (vector_game) {
		compute_game_variable_vector(&OSD_REQ);

		if (!compute_mode())
			return -1;
	}

	return 0;
}

int osd_create_display(int width,int height)
{
	OSD_REQ.width = width;
	OSD_REQ.height = height;
	return 0;
}

/* center image inside the display based on the visual area */
void osd_set_visible_area(int min_x,int max_x,int min_y,int max_y)
{
	logerror("set visible area %d-%d %d-%d\n",min_x,max_x,min_y,max_y);

	OSD_REQ.min_x = min_x;
	OSD_REQ.min_y = min_y;
	OSD_REQ.max_x = max_x;
	OSD_REQ.max_y = max_y;
}

int osd_allocate_colors(unsigned int totalcolors, const UINT8 *palette, UINT16 *pens, int modifiable, const UINT8 *debug_palette, UINT16 *debug_pens)
{
	OSD_REQ.totalcolors = totalcolors;
	OSD_REQ.palette = palette;
	OSD_REQ.pens = pens;
	OSD_REQ.modifiable = modifiable;

	/* Compute late the video mode for raster game */
	/* In this way the video system can choice the video mode knowing the exact size of the game bitmap */
	if (!vector_game) {
		compute_game_variable_raster(&OSD_REQ);

		if (!compute_mode())
			return -1;
	}

	if (!activate_video_configuration())
		return -1;

	if (!compute_sync())
		return -1;

	compute_video_variable();

	if (!compute_palette(&OSD_REQ))
		return -1;

	compute_changeable();

	return 0;
}


void osd_close_display(void)
{
	free_all();
}

void osd_modify_pen(int pen,unsigned char red, unsigned char green, unsigned char blue)
{
	if (current_palette[pen].red != red || current_palette[pen].green != green || current_palette[pen].blue != blue)
	{
		current_palette[pen].red = red;
		current_palette[pen].green = green;
		current_palette[pen].blue = blue;

		palette_dirty_color[pen] = 1;
		palette_dirty = 1;
	}
}

void osd_get_pen(int pen,unsigned char *red, unsigned char *green, unsigned char *blue)
{
	if (!current_palette)
	{
		*red = video_red_get_approx(pen);
		*green = video_green_get_approx(pen);
		*blue = video_blue_get_approx(pen);
	}
	else
	{
		*red = current_palette[pen].red;
		*green = current_palette[pen].green;
		*blue = current_palette[pen].blue;
	}
}


void osd_set_gamma(float _gamma)
{
	int i;

	osd_gamma_correction = _gamma;

	bright_dirty = 1;

	if (!palette_rgb) {
		for (i = 0;i < palette_total;i++)
		{
			if (i != Machine->uifont->colortable[1])        /* don't touch the user interface text */
				palette_dirty_color[i] = 1;
		}
		palette_dirty = 1;
	}
}

float osd_get_gamma(void)
{
	return osd_gamma_correction;
}

/* brightess = percentage 0-100% */
void osd_set_brightness(int _brightness)
{
	int i;

	brightness = _brightness;

	for (i = 0;i < palette_total;i++)
		palette_dirty_color[i] = 1;
	palette_dirty = 1;
	bright_dirty = 1;
}

int osd_get_brightness(void)
{
	return brightness;
}

void osd_mark_dirty(int xmin, int ymin, int xmax, int ymax)
{
	if (dirty_effective)
	{
		int posx = xmin;
		int posy = ymin;
		int sizex = xmax - xmin + 1;
		int sizey = ymax - ymin + 1;

		if (blit_orientation & ORIENTATION_SWAP_XY) {
			int temp = posx;
			posx = posy;
			posy = temp;
			temp = sizex;
			sizex = sizey;
			sizey = temp;
		}

		if (blit_orientation & ORIENTATION_FLIP_X) {
			posx = game_width - sizex - posx;
		}

		if (blit_orientation & ORIENTATION_FLIP_Y) {
			posy = game_height - sizey - posy;
		}

		/* reference at the visible part of the game screen */
		posx -= game_visible_posx;
		posy -= game_visible_posy;

		dirty_set_bitmap_region_clip(posx,posy,posx+sizex-1,posy+sizey-1);
	}
}

void osd_mark_dirty_point(int posx, int posy)
{
	if (dirty_effective)
	{
		if (blit_orientation & ORIENTATION_SWAP_XY) {
			int temp = posx;
			posx = posy;
			posy = temp;
		}

		if (blit_orientation & ORIENTATION_FLIP_X) {
			posx = game_width - 1 - posx;
		}

		if (blit_orientation & ORIENTATION_FLIP_Y) {
			posy = game_height - 1 - posy;
		}

		/* reference at the visible part of the game screen */
		posx -= game_visible_posx;
		posy -= game_visible_posy;

		dirty_set(posx / DIRTY_DELTA_X, posy / DIRTY_DELTA_Y);
	}
}

void osd_save_sound_stop(void)
{
	if (sound_record_time) {
		unsigned saved;
		if (sound_record_stop(&saved)==0) {
			/* recording stopped */
			usrintf_showmessage("Sound saved, %d seconds", (unsigned)(saved / Machine->sample_rate));
		}
	}
}

void osd_save_sound_start(void)
{
	if (sound_record_time) {
		char path[256];
		extern char *screenshotdir;
		static int snapno;

		/* if got an error start the recording */
		sprintf(path,"%s/%.8s.wav", screenshotdir, Machine->gamedrv->name);
		if (access(path,F_OK)==0)
		{
			do
			{
				/* otherwise use "nameNNNN.png" */
				sprintf(path,"%s/%.4s%04d.wav", screenshotdir, Machine->gamedrv->name,snapno++);
			} while (access(path,F_OK)==0);
		}

		usrintf_showmessage("Start sound recording");

		sound_record_cancel(); /* ignore error */

		sound_record_start(path, sound_record_time*Machine->sample_rate, Machine->sample_rate, (Machine->drv->sound_attributes & SOUND_SUPPORTS_STEREO)!=0);
	}
}

void osd_save_snapshot(struct osd_bitmap *bitmap)
{
	save_screen_snapshot(bitmap);
	usrintf_showmessage("Snapshot saved");
}

void osd_pause(int paused)
{
	int i;

	if (paused) brightness_paused_adjust = 0.65;
	else brightness_paused_adjust = 1.0;

	for (i = 0;i < palette_total;i++)
		palette_dirty_color[i] = 1;
	palette_dirty = 1;
	bright_dirty = 1;       
}

void osd_reset(void)
{
	/* inizialize the turbo frame counter */
	turbo_frame_counter = 0;
}

void osd_debugger_focus(int debugger_has_focus)
{
}

/***************************************************************************/
/* Update */

static int change_check_video_configuration(const video_config* active_cfg, const video_config* cfg)
{
	if (active_cfg == cfg)
		return 1;

	if (!check_video_configuration(cfg))
		return 0;

	/* vector games don't support mode with different size because */
	/* isn't possible to resize the vector bitmap */
	if (vector_game) {
		if (video_mode_size_x(cfg->mode) != video_mode_size_x(active_cfg->mode))
			return 0;
		if (video_mode_size_y(cfg->mode) != video_mode_size_y(active_cfg->mode))
			return 0;
	}

	/* check that the rgb nibble is the same */
	if (palette_rgb) {
		/* TODO check the exact match of rgb nibble */
		if (video_mode_bits_per_pixel(cfg->mode) != video_mode_bits_per_pixel(active_cfg->mode))
			return 0;
	}

	/* if vsync is active no mode change is possible for the different frame rate */
	if (video_vsync_request) {
		if (active_cfg->mode != cfg->mode) {
			return 0;
		}
	}

	return 1;
}

static __inline__ void update_resolution(void)
{
	int modify = 0;
	video_config* cfg = gfx_cfg;

	if (input_ui_pressed(IPT_UI_MODE_NEXT)) {
		++cfg;
		while (cfg != gfx_cfg_map + gfx_cfg_mac && !change_check_video_configuration(gfx_cfg,cfg))
			++cfg;
		modify = 1;
	} else if (input_ui_pressed(IPT_UI_MODE_PRED)) {
		--cfg;
		while (cfg != gfx_cfg_map - 1 && !change_check_video_configuration(gfx_cfg,cfg))
			--cfg;
		modify = 1;
	}

	if (modify) {
		if (cfg != gfx_cfg_map - 1 && cfg != gfx_cfg_map + gfx_cfg_mac) {
			strcpy(resolution_request,video_configuration_name(cfg));
			change_state();
		}

		/* show at screen the new configuration name */
		usrintf_showmessage(video_configuration_name(gfx_cfg));
	}
}

static __inline__ void update_skip(void)
{
	int modify = 0;

	if (input_ui_pressed_repeat(IPT_UI_PAN_RIGHT,1))
	{
		if (game_visible_posx + gfx_cfg->game_visible_width + game_visible_posx_increment <= game_used_posx + game_used_width)
		{
			game_visible_posx += game_visible_posx_increment;
			modify = 1;
		}
	}

	if (input_ui_pressed_repeat(IPT_UI_PAN_LEFT,1))
	{
		if (game_visible_posx >= game_used_posx + game_visible_posx_increment)
		{
			game_visible_posx -= game_visible_posx_increment;
			modify = 1;
		}
	}

	if (input_ui_pressed_repeat(IPT_UI_PAN_UP,1))
	{
		if (game_visible_posy + gfx_cfg->game_visible_height < game_used_posy + game_used_height)
		{
			game_visible_posy++;
			modify = 1;
		}
	}

	if (input_ui_pressed_repeat(IPT_UI_PAN_DOWN,1))
	{
		if (game_visible_posy > game_used_posy)
		{
			game_visible_posy--;
			modify = 1;
		}
	}

	if (modify) {
		skipcolumns_request = game_visible_posx - game_used_posx;
		skiplines_request = game_visible_posy - game_used_posy;
		change_skip();
	}
}

int osd_skip_this_frame(void)
{
	static const int skiptable[FRAMESKIP_LEVELS][FRAMESKIP_LEVELS] =
	{
		{ 0,0,0,0,0,0,0,0,0,0,0,0 },
		{ 0,0,0,0,0,0,0,0,0,0,0,1 },
		{ 0,0,0,0,0,1,0,0,0,0,0,1 },
		{ 0,0,0,1,0,0,0,1,0,0,0,1 },
		{ 0,0,1,0,0,1,0,0,1,0,0,1 },
		{ 0,1,0,0,1,0,1,0,0,1,0,1 },
		{ 0,1,0,1,0,1,0,1,0,1,0,1 },
		{ 0,1,0,1,1,0,1,0,1,1,0,1 },
		{ 0,1,1,0,1,1,0,1,1,0,1,1 },
		{ 0,1,1,1,0,1,1,1,0,1,1,1 },
		{ 0,1,1,1,1,1,0,1,1,1,1,1 },
		{ 0,1,1,1,1,1,1,1,1,1,1,1 }
	};

	return skiptable[frameskip][frameskip_counter];
}

static __inline__ void update_blit(unsigned dst_x, unsigned dst_y, unsigned dst_dx, unsigned dst_dy, void* src, unsigned src_dx, unsigned src_dy, int src_dw, int src_dp)
{
	if (!palette_lookup) {
		video_stretch(dst_x, dst_y, dst_dx, dst_dy, src, src_dx, src_dy, src_dw, src_dp, blit_combine_effective);
	} else {
		switch (game_bytes_per_pixel) {
			case 1 :
				video_stretch_palette8(dst_x, dst_y, dst_dx, dst_dy, src, src_dx, src_dy, src_dw, src_dp, palette_lookup, blit_combine_effective);
				break;
			case 2 :
				video_stretch_palette16(dst_x, dst_y, dst_dx, dst_dy, src, src_dx, src_dy, src_dw, src_dp, palette_lookup, blit_combine_effective);
				break;
		}
	}
}

static __inline__ void update_pipeline(void) {
	if (!blit_pipeline_flag) {
		if (!palette_lookup) {
			video_stretch_pipeline_init(&blit_pipeline, gfx_cfg->screen_visible_width, gfx_cfg->screen_visible_height, gfx_cfg->game_visible_width, gfx_cfg->game_visible_height, blit_dw, blit_dp, blit_combine_effective);
		} else {
			switch (game_bytes_per_pixel) {
				case 1 :
					video_stretch_palette8_pipeline_init(&blit_pipeline, gfx_cfg->screen_visible_width, gfx_cfg->screen_visible_height, gfx_cfg->game_visible_width, gfx_cfg->game_visible_height, blit_dw, blit_dp, palette_lookup, blit_combine_effective);
					break;
				case 2 :
					video_stretch_palette16_pipeline_init(&blit_pipeline, gfx_cfg->screen_visible_width, gfx_cfg->screen_visible_height, gfx_cfg->game_visible_width, gfx_cfg->game_visible_height, blit_dw, blit_dp,  palette_lookup, blit_combine_effective);
					break;
			}
		}

		{
			int i;
			const struct video_stage_horz_struct* stage;

			logerror("Video pipeline:\n");
			for(i=1,stage=video_pipeline_begin(&blit_pipeline);stage!=video_pipeline_end(&blit_pipeline);++stage,++i) {
				if (stage == video_pipeline_pivot(&blit_pipeline)) {
					logerror("\t(%d) %s\n",i,pipe_name(video_pipeline_vert(&blit_pipeline)->type));
					++i;
				}
				logerror("\t(%d) %s, pixel %d bytes, step %d bytes\n",i,pipe_name(stage->type),stage->sbpp,stage->sdp);
			}
		}

		blit_pipeline_flag = 1;
	}
}

static __inline__ void update_screen(struct osd_bitmap *bitmap) {
	unsigned dst_x;
	unsigned dst_y;
	int src_dw;
	int src_dp;
	int src_offset;

	video_update_start();

	/* screen position */
	dst_x = blit_x + video_update_x();
	dst_y = blit_y + video_update_y();

	/* source increment */
	src_dw = blit_dw;
	src_dp = blit_dp;

	/* compute the source pointer */
	src_offset = blit_offset + game_visible_posy * src_dw + game_visible_posx * src_dp;

	update_pipeline();

	/* frational stretch_request don't support dirty */
	if (dirty_effective) {
		/* number of dirty block */
		unsigned bl_dx = (gfx_cfg->game_visible_width + DIRTY_DELTA_X - 1) / DIRTY_DELTA_X;
		unsigned bl_dy = (gfx_cfg->game_visible_height + DIRTY_DELTA_Y - 1) / DIRTY_DELTA_Y;

		/* for every dirty row */
		unsigned bl_y = 0;

		assert( screen_stretch!=SCREEN_STRETCH_FRACTIONAL && blit_combine_effective==COMBINE_NONE );

		while (bl_y<bl_dy) {
			unsigned scr_y;
			unsigned scr_dy;
			unsigned bl_x;

			/* bitmap coordinate */
			unsigned y = bl_y * DIRTY_DELTA_Y;
			unsigned dy = DIRTY_DELTA_Y;
			if (y+dy > gfx_cfg->game_visible_height)
				dy = gfx_cfg->game_visible_height - y;

			/* screen coordinate */
			scr_y = dst_y + y * screen_stretch_factory;
			scr_dy = dy * screen_stretch_factory;

			/* for every dirty column */
			bl_x = 0;
			while (bl_x<bl_dx) {
				unsigned bl_x_end;
				/* search first block dirty */
				while (bl_x<bl_dx && !is_dirty(bl_x,bl_y))
					++bl_x;
				/* search next block not dirty */
				bl_x_end = bl_x;
				while (bl_x_end<bl_dx && is_dirty(bl_x_end,bl_y))
					++bl_x_end;
				if (bl_x!=bl_x_end) {
					char* src;
					unsigned scr_x;
					unsigned scr_dx;

					/* bitmap coordinate */
					unsigned x = bl_x * DIRTY_DELTA_X;
					unsigned dx = (bl_x_end - bl_x) * DIRTY_DELTA_X;
					if (x+dx > gfx_cfg->game_visible_width)
						dx = gfx_cfg->game_visible_width - x;

					/* screen coordinate */
					scr_x = dst_x + x * screen_stretch_factorx;
					scr_dx = dx * screen_stretch_factorx;

					src = (char*)bitmap->line[0] + y * src_dw + x * src_dp;

					update_blit(
						scr_x, scr_y,
						scr_dx, scr_dy,
						src + src_offset,
						dx,dy,
						src_dw,
						src_dp
					);
				}
				/* next block */
				bl_x = bl_x_end;
			}
			/* next row */
			++bl_y;
		}
	} else {
		uint8* src = (uint8*)bitmap->line[0];
		video_blit_pipeline(&blit_pipeline, dst_x, dst_y, src + src_offset);
	}

#ifdef __SNAPSHOT__
	if (input_ui_pressed_repeat(IPT_UI_SNAPSHOT,1)) {
		static int snapshot_num = 0;
		char snapshot_name[256];
		sprintf(snapshot_name,"snap%d.bmp",snapshot_num);
		video_snapshot_save(snapshot_name,video_update_x(),video_update_y());
		++snapshot_num;
	}
#endif

	video_update_stop(0);
}

static __inline__ void update_bright(void) {
	if (bright_dirty)
	{
		int i;
		double factor = 255.0 * brightness * brightness_paused_adjust / 100.0;
		bright_dirty = 0;
		for (i = 0;i < 256;i++)
		{
			double rate = factor *  pow(i / 255.0, 1 / osd_gamma_correction);
			if (rate>255)
				rate = 255;
			bright_lookup[i] = rate;
		}

		if (!palette_rgb) {
			palette_dirty = 1;
			for (i = 0;i < palette_total;i++)
				palette_dirty_color[i] = 1;
		}
	}
}

static __inline__ void update_palette(void) {
	if (palette_dirty && !palette_rgb)
	{
		int i;
		palette_dirty = 0;
		for (i = 0;i < palette_total;i++)
		{
			if (palette_dirty_color[i])
			{
				video_color adjusted_palette;
				
				palette_dirty_color[i] = 0;

				adjusted_palette.red = current_palette[i].red;
				adjusted_palette.green = current_palette[i].green;
				adjusted_palette.blue = current_palette[i].blue;
				if (i != Machine->uifont->colortable[1]) /* don't adjust the user interface text */
				{
					adjusted_palette.red = bright_lookup[adjusted_palette.red];
					adjusted_palette.green = bright_lookup[adjusted_palette.green];
					adjusted_palette.blue = bright_lookup[adjusted_palette.blue];
				}

				/* update the palette */
				if (!palette_lookup)
				{
					/* hardware */
					/* note: trying to concatenate palette update */
					/* generate flickering!, one color at time is ok! */
					video_palette_set(&adjusted_palette,i,1,0);
				} else {
					/* software */
					video_rgb rgb;
					video_rgb_make(&rgb,adjusted_palette.red,adjusted_palette.green,adjusted_palette.blue);
					palette_lookup[i] = rgb;
				}
			}
		}

		if (palette_lookup && dirty_effective)
			dirty_set_all();
	}
}

/* Update the display. */
void osd_update_video_and_audio(struct osd_bitmap *game_bitmap,struct osd_bitmap *debug_bitmap,int leds_status)
{
	static const int waittable[FRAMESKIP_LEVELS][FRAMESKIP_LEVELS] =
	{
		{ 1,1,1,1,1,1,1,1,1,1,1,1 },
		{ 2,1,1,1,1,1,1,1,1,1,1,0 },
		{ 2,1,1,1,1,0,2,1,1,1,1,0 },
		{ 2,1,1,0,2,1,1,0,2,1,1,0 },
		{ 2,1,0,2,1,0,2,1,0,2,1,0 },
		{ 2,0,2,1,0,2,0,2,1,0,2,0 },
		{ 2,0,2,0,2,0,2,0,2,0,2,0 },
		{ 2,0,2,0,0,3,0,2,0,0,3,0 },
		{ 3,0,0,3,0,0,3,0,0,3,0,0 },
		{ 4,0,0,0,4,0,0,0,4,0,0,0 },
		{ 6,0,0,0,0,0,6,0,0,0,0,0 },
		{12,0,0,0,0,0,0,0,0,0,0,0 }
	};
	static int showfps,showfpstemp;
	TICKER curr;
	static TICKER prev_measure,this_frame_base,prev;
	static int speed = 100;
	static int vups,vfcount;
	int already_synced;
	struct osd_bitmap *bitmap;
	static int leds_old = -1;
	int turbo_active; /* turbo active flag */
	int turbo_video_fps; /* adjusted fps */

	/* turbo */
	++turbo_frame_counter;
	if (turbo_frame_counter < turbo_startup_time * video_fps) {
		turbo_active = 1;
		turbo_video_fps = video_fps * turbo_startup_speed;
	} else if (seq_pressed(input_port_type_seq(IPT_UI_TURBO))) {
		turbo_active = 1;
		turbo_video_fps = video_fps * turbo_speed;
	} else {
		turbo_active = 0;
		turbo_video_fps = video_fps;
	}

	if (leds_old != leds_status)
	{
		int i;
		static const int led_flags[3] =
		{
			KB_NUMLOCK_FLAG,
			KB_CAPSLOCK_FLAG,
			KB_SCROLOCK_FLAG
		};

		leds_old = leds_status;

		i = 0;
		if (leds_status & 1) i |= led_flags[0];
		if (leds_status & 2) i |= led_flags[1];
		if (leds_status & 4) i |= led_flags[2];
		set_leds(i);
	}

	bitmap = game_bitmap;

	if (warming_up)
	{
		/* first time through, initialize timer */
		prev_measure = ticker() - FRAMESKIP_LEVELS * TICKS_PER_SEC/turbo_video_fps;
		warming_up = 0;
	}

	if (frameskip_counter == 0)
		this_frame_base = prev_measure + FRAMESKIP_LEVELS * TICKS_PER_SEC/turbo_video_fps;

	if (throttle)
	{
		static TICKER last;

		/* if too much time has passed since last sound update, disable throttling */
		/* temporarily - we wouldn't be able to keep synch anyway. */
		curr = ticker();
		if (turbo_active || (curr - last) > 2*TICKS_PER_SEC / video_fps)
			throttle = 0;
		last = curr;

		already_synced = msdos_update_audio() && !turbo_active;

		throttle = 1;
	}
	else
		already_synced = msdos_update_audio();

	if (osd_skip_this_frame() == 0)
	{
		if (showfpstemp)
		{
			showfpstemp--;
			if (showfps == 0 && showfpstemp == 0)
				schedule_full_refresh();
		}

		if (input_ui_pressed(IPT_UI_SHOW_FPS))
		{
			if (showfpstemp)
			{
				showfpstemp = 0;
				schedule_full_refresh();
			}
			else
			{
				showfps ^= 1;
				if (showfps == 0)
				{
					schedule_full_refresh();
				}
			}
		}

		/* now wait until it's time to update the screen */
		if (throttle)
		{
			profiler_mark(PROFILER_IDLE);
			if (video_vsync_request && !turbo_active)
			{
				static TICKER last;

				do
				{
					video_wait_vsync();
					curr = ticker();
				} while (TICKS_PER_SEC / (curr - last) > video_fps * 11 /10);

				last = curr;
			}
			else
			{
				TICKER target;


				/* wait for video sync but use normal throttling */
				if (wait_vsync_request && !turbo_active)
					video_wait_vsync();

				curr = ticker();

				if (already_synced == 0)
				{
				/* wait only if the audio update hasn't synced us already */

					target = this_frame_base +
							frameskip_counter * TICKS_PER_SEC/turbo_video_fps;

					if (curr - target < 0)
					{
						do
						{
							curr = ticker();
						} while (curr - target < 0);
					}
				}

			}
			profiler_mark(PROFILER_END);
		}
		else curr = ticker();


		/* for the FPS average calculation */
		if (++frames_displayed == FRAMES_TO_SKIP)
			start_time = curr;
		else
			end_time = curr;


		if (frameskip_counter == 0)
		{
			int divdr;

			divdr = video_fps  * (curr - prev_measure) / (100 * FRAMESKIP_LEVELS);
			speed = (TICKS_PER_SEC + divdr/2) / divdr;

			prev_measure = curr;
		}

		prev = curr;

		vfcount += waittable[frameskip][frameskip_counter];
		if (vfcount >= video_fps )
		{
			extern int vector_updates; /* avgdvg_go()'s per Mame frame, should be 1 */


			vfcount = 0;
			vups = vector_updates;
			vector_updates = 0;
		}

		if (showfps || showfpstemp)
		{
			int fps;
			char buf[30];
			int divdr;

			divdr = 100 * FRAMESKIP_LEVELS;
			fps = (video_fps * (FRAMESKIP_LEVELS - frameskip) * speed + (divdr / 2)) / divdr;
			sprintf(buf,"%s%2d%4d%%%4d/%d fps",autoframeskip?"auto":"fskp",frameskip,speed,fps,(int)(video_fps+0.5));
			ui_text(bitmap,buf,Machine->uiwidth-strlen(buf)*Machine->uifontwidth,0);
			if (vector_game)
			{
				sprintf(buf," %d vector updates",vups);
				ui_text(bitmap,buf,Machine->uiwidth-strlen(buf)*Machine->uifontwidth,Machine->uifontheight);
			}
		}

		/* update the bright lookup */
		update_bright();
		
		/* update the palette lookup */
		update_palette();

		/* copy the bitmap to screen memory */
		profiler_mark(PROFILER_BLIT);
		update_screen(bitmap);
		profiler_mark(PROFILER_END);

		if (dirty_effective)
		{
			dirty_shift();
		}

		if (throttle && autoframeskip && frameskip_counter == 0)
		{
			static int frameskipadjust;
			int adjspeed;

			/* adjust speed to video refresh rate if vsync is on */
			adjspeed = speed * video_fps / vsync_frame_rate * video_fps / turbo_video_fps;

			if (adjspeed >= 100)
			{
				frameskipadjust++;
				if (frameskipadjust >= 3)
				{
					frameskipadjust = 0;
					if (frameskip > 0) frameskip--;
				}
			}
			else
			{
				if (adjspeed < 80)
					frameskipadjust -= (90 - adjspeed) / 5;
				else
				{
					/* don't push frameskip too far if we are close to 100% speed */
					if (frameskip < 8)
						frameskipadjust--;
				}

				while (frameskipadjust <= -2)
				{
					frameskipadjust += 2;
					if (frameskip < FRAMESKIP_LEVELS-1) frameskip++;
				}
			}
		}
	}

	update_resolution();

	update_skip();

	if (input_ui_pressed(IPT_UI_FRAMESKIP_INC))
	{
		if (autoframeskip)
		{
			autoframeskip = 0;
			frameskip = 0;
		}
		else
		{
			if (frameskip == FRAMESKIP_LEVELS-1)
			{
				frameskip = 0;
				autoframeskip = 1;
			}
			else
				frameskip++;
		}

		if (showfps == 0)
			showfpstemp = 2*video_fps;

		/* reset the frame counter every time the frameskip key is pressed, so */
		/* we'll measure the average FPS on a consistent status. */
		frames_displayed = 0;
	}

	if (input_ui_pressed(IPT_UI_FRAMESKIP_DEC))
	{
		if (autoframeskip)
		{
			autoframeskip = 0;
			frameskip = FRAMESKIP_LEVELS-1;
		}
		else
		{
			if (frameskip == 0)
				autoframeskip = 1;
			else
				frameskip--;
		}

		if (showfps == 0)
			showfpstemp = 2*video_fps;

		/* reset the frame counter every time the frameskip key is pressed, so */
		/* we'll measure the average FPS on a consistent status. */
		frames_displayed = 0;
	}

	if (input_ui_pressed(IPT_UI_THROTTLE))
	{
		throttle ^= 1;

		/* reset the frame counter every time the throttle key is pressed, so */
		/* we'll measure the average FPS on a consistent status. */
		frames_displayed = 0;
	}


	frameskip_counter = (frameskip_counter + 1) % FRAMESKIP_LEVELS;

	poll_joysticks();
}

/***************************************************************************/
/* Menu */

static int video_mode_menu(struct osd_bitmap *bitmap, int selected)
{
	const char *menu_item[GFX_CFG_MAX + 2];
	char flag[GFX_CFG_MAX + 2];
	video_config* entry[GFX_CFG_MAX + 2];
	int i;
	int sel;
	int total;

	sel = selected - 1;

	total = 0;

	menu_item[total] = strdup("Auto");
	entry[total] = 0;
	flag[total] = strcmp("auto",resolution_request)==0;
	++total;

	for(i=gfx_cfg_mac-1;i>=0;--i) {
		video_config* cfg = gfx_cfg_map + i;
		if (change_check_video_configuration(gfx_cfg,cfg)) {
			entry[total] = cfg;
			menu_item[total] = strdup(video_configuration_desc(cfg));
			flag[total] = strcmp(video_configuration_name(cfg),resolution_request)==0;
			++total;
		}
	}

	menu_item[total] = ui_getstring (UI_returntomain);
	flag[total] = 0;
	++total;

	menu_item[total] = 0;
	flag[total] = 0;

	ui_displaymenu(bitmap,menu_item,0,flag,sel,0);

	if (input_ui_pressed_repeat(IPT_UI_DOWN,8))
	{
		sel = (sel + 1) % total;
	}

	if (input_ui_pressed_repeat(IPT_UI_UP,8))
	{
		sel = (sel + total - 1) % total;
	}

	if (input_ui_pressed(IPT_UI_SELECT))
	{
		if (sel == total - 1) sel = -1;
		else if (sel == 0) {
			strcpy(resolution_request,"auto");
			change_state();

			/* show at screen the new configuration name */
			usrintf_showmessage(video_configuration_name(gfx_cfg));

			/* tell updatescreen() to clean after us */
			schedule_full_refresh();
		} else {
			strcpy(resolution_request,video_configuration_name(entry[sel]));
			change_state();
			
			/* tell updatescreen() to clean after us */
			schedule_full_refresh();
		}
	}

	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CONFIGURE))
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		/* tell updatescreen() to clean after us */
		schedule_full_refresh();
	}

	for(i=0;i<total-1;++i) {
		free((char*)menu_item[i]);
	}

	return sel + 1;
}

char* get_game_resolution(const struct GameDriver* driver) {
	static char internal_resolution[32];
	const struct MachineDriver* machine = driver->drv;
	if ((machine->video_attributes & VIDEO_TYPE_VECTOR) == 0) {
		unsigned dx;
		unsigned dy;
		if (driver->flags & ORIENTATION_SWAP_XY)
		{
			dx = machine->default_visible_area.max_y - machine->default_visible_area.min_y + 1;
			dy = machine->default_visible_area.max_x - machine->default_visible_area.min_x + 1;
		}
		else
		{
			dx = machine->default_visible_area.max_x - machine->default_visible_area.min_x + 1;
			dy = machine->default_visible_area.max_y - machine->default_visible_area.min_y + 1;
		}
		sprintf(internal_resolution,"%dx%d",dx,dy);
	} else {
		strcpy(internal_resolution,"vector");
	}
	return internal_resolution;
}

int osd_menu(struct osd_bitmap *bitmap, int selected)
{
	const char *menu_item[16];
	const char *menu_subitem[16];
	char flag[16];
	int sel;
	int total;
	int arrowize;
	static int subsel = 0;

	int resolution_index;
	int stretch_index;
	int wait_index;
	int combine_index;
	int save_game_index;
	int save_resolution_index;
	int save_all_index;

	sel = selected - 1;

	if (subsel)
	{
		int ret = video_mode_menu(bitmap,subsel);
		switch (ret) {
			case -1 : return -1;
			case 0 : subsel = 0; break;
			default: subsel = ret; break;
		}
		return sel + 1;
	}

	total = 0;

	resolution_index = total;
	menu_item[total] = "Video config";
	if (strcmp(resolution_request,"auto")==0)
		menu_subitem[total] = "auto";
	else
		menu_subitem[total] = "custom";
	flag[total] = 0;
	++total;

	if (!vector_game) {
		if (strcmp(resolution_request,"auto")==0) {
			stretch_index = total;
			switch (video_configuration_stretch(gfx_cfg)) {
				case STRETCH_NONE : menu_item[total] = "Resize [no]"; break;
				case STRETCH_INTEGER_XY : menu_item[total] = "Resize [integer]"; break;
				case STRETCH_INTEGER_X_FRACTIONAL_Y : menu_item[total] = "Resize [mixed]"; break;
				case STRETCH_FRACTIONAL_XY : menu_item[total] = "Resize [fractional]"; break;
			}
			switch (stretch_request) {
				case STRETCH_NONE : menu_subitem[total] = "no"; break;
				case STRETCH_INTEGER_XY : menu_subitem[total] = "integer"; break;
				case STRETCH_INTEGER_X_FRACTIONAL_Y : menu_subitem[total] = "mixed"; break;
				case STRETCH_FRACTIONAL_XY : menu_subitem[total] = "fractional"; break;
			}
			flag[total] = 0;
			++total;
		} else {
			stretch_index = -1;
			switch (video_configuration_stretch(gfx_cfg)) {
				case STRETCH_NONE : menu_item[total] = "Resize [no]"; break;
				case STRETCH_INTEGER_XY : menu_item[total] = "Resize [integer]"; break;
				case STRETCH_INTEGER_X_FRACTIONAL_Y : menu_item[total] = "Resize [mixed]"; break;
				case STRETCH_FRACTIONAL_XY : menu_item[total] = "Resize [fractional]"; break;
			}
			menu_subitem[total] = "custom";
			flag[total] = 0;
			++total;
		}
	} else {
		stretch_index = -1;
	}

	combine_index = total;
	switch (blit_combine_effective) {
		case COMBINE_NONE : menu_item[total] = "Effect [no]"; break;
		case COMBINE_MAX : menu_item[total] = "Effect [max]"; break;
		case COMBINE_MEAN : menu_item[total] = "Effect [mean]"; break;
		case COMBINE_FILTER : menu_item[total] = "Effect [filter]"; break;
		case COMBINE_FILTERX : menu_item[total] = "Effect [filterx]"; break;
		case COMBINE_FILTERY : menu_item[total] = "Effect [filtery]"; break;
	}
	switch (blit_combine_request) {
		case COMBINE_AUTO : menu_subitem[total] = "auto"; break;
		case COMBINE_NONE : menu_subitem[total] = "no"; break;
		case COMBINE_MAX : menu_subitem[total] = "max"; break;
		case COMBINE_MEAN : menu_subitem[total] = "mean"; break;
		case COMBINE_FILTER : menu_subitem[total] = "filter"; break;
		case COMBINE_FILTERX : menu_subitem[total] = "filterx"; break;
		case COMBINE_FILTERY : menu_subitem[total] = "filtery"; break;
	}
	flag[total] = 0;
	++total;

	wait_index = total;
	switch (wait_vsync_effective) {
		case 0 : menu_item[total] = "WaitVsync [no]"; break;
		case 1 : menu_item[total] = "WaitVsync [yes]"; break;
	}
	switch (wait_vsync_request) {
		case 0 : menu_subitem[total] = "no"; break;
		case 1 : menu_subitem[total] = "yes"; break;
	}
	flag[total] = 0;
	++total;

	save_game_index = total;
	menu_item[total] = "Save for this game";
	menu_subitem[total] = 0;
	flag[total] = 0;
	++total;

	save_resolution_index = total;
	if (vector_game)
		menu_item[total] = "Save for all vector games";
	else
		menu_item[total] = "Save for this resolution";
	menu_subitem[total] = 0;
	flag[total] = 0;
	++total;

	if (!vector_game) {
		save_all_index = total;
		menu_item[total] = "Save for all games";
		menu_subitem[total] = 0;
		flag[total] = 0;
		++total;
	} else {
		save_all_index = -1;
	}

	menu_item[total] = ui_getstring (UI_returntomain);
	menu_subitem[total] = 0;
	flag[total] = 0;
	++total;

	menu_item[total] = 0;
	menu_subitem[total] = 0;
	flag[total] = 0;

	if (sel == wait_index || sel == stretch_index || sel == combine_index)
		arrowize = 3;
	else
		arrowize = 0;

	ui_displaymenu(bitmap,menu_item,menu_subitem,flag,sel,arrowize);

	if (input_ui_pressed_repeat(IPT_UI_DOWN,8))
	{
		sel = (sel + 1) % total;
	}

	if (input_ui_pressed_repeat(IPT_UI_UP,8))
	{
		sel = (sel + total - 1) % total;
	}

	if (input_ui_pressed(IPT_UI_SELECT))
	{
		if (sel == total - 1) {
			sel = -1;
		} else if (sel == resolution_index) {
			subsel = 1;
			/* tell updatescreen() to clean after us */
			schedule_full_refresh();
		} else if (sel == save_game_index) {
			save_state(Machine->gamedrv->name);
		} else if (sel == save_resolution_index) {
			save_state(get_game_resolution(Machine->gamedrv));
		} else if (sel == save_all_index) {
			save_state("config");
		}
	}

	if (input_ui_pressed_repeat(IPT_UI_RIGHT,8))
	{
		if (sel == combine_index) {
			switch (blit_combine_request) {
				case COMBINE_AUTO : blit_combine_request = COMBINE_NONE; break;
				case COMBINE_NONE : blit_combine_request = COMBINE_MAX; break;
				case COMBINE_MAX : blit_combine_request = COMBINE_MEAN; break;
				case COMBINE_MEAN : blit_combine_request = COMBINE_FILTER; break;
				case COMBINE_FILTER : blit_combine_request = COMBINE_FILTERX; break;
				case COMBINE_FILTERX : blit_combine_request = COMBINE_FILTERY; break;
				case COMBINE_FILTERY : blit_combine_request = COMBINE_AUTO; break;
			}
			change_state();
			/* tell updatescreen() to clean after us */
			schedule_full_refresh();
		} else if (sel == wait_index) {
			switch (wait_vsync_request) {
				case 0 : wait_vsync_request = 1; break;
				case 1 : wait_vsync_request = 0; break;
			}
			change_state();
			/* tell updatescreen() to clean after us */
			schedule_full_refresh();
		} else if (sel == stretch_index) {
			switch (stretch_request) {
				case STRETCH_NONE : stretch_request = STRETCH_FRACTIONAL_XY; break;
				case STRETCH_INTEGER_XY :  stretch_request = STRETCH_NONE; break;
				case STRETCH_INTEGER_X_FRACTIONAL_Y : stretch_request = STRETCH_INTEGER_XY; break;
				case STRETCH_FRACTIONAL_XY : stretch_request = STRETCH_INTEGER_X_FRACTIONAL_Y; break;
			}
			strcpy(resolution_request,"auto");
			change_state();
			/* tell updatescreen() to clean after us */
			schedule_full_refresh();
		}
	}

	if (input_ui_pressed_repeat(IPT_UI_LEFT,8))
	{
		if (sel == combine_index) {
			switch (blit_combine_request) {
				case COMBINE_AUTO : blit_combine_request = COMBINE_FILTERY; break;
				case COMBINE_NONE : blit_combine_request = COMBINE_AUTO; break;
				case COMBINE_MAX : blit_combine_request = COMBINE_NONE; break;
				case COMBINE_MEAN : blit_combine_request = COMBINE_MAX; break;
				case COMBINE_FILTER : blit_combine_request = COMBINE_MEAN; break;
				case COMBINE_FILTERX : blit_combine_request = COMBINE_FILTER; break;
				case COMBINE_FILTERY : blit_combine_request = COMBINE_FILTERX; break;
			}
			change_state();
			/* tell updatescreen() to clean after us */
			schedule_full_refresh();
		} else if (sel == wait_index) {
			switch (wait_vsync_request) {
				case 0 : wait_vsync_request = 1; break;
				case 1 : wait_vsync_request = 0; break;
			}
			change_state();
			/* tell updatescreen() to clean after us */
			schedule_full_refresh();
		} else if (sel == stretch_index) {
			switch (stretch_request) {
				case STRETCH_NONE : stretch_request = STRETCH_INTEGER_XY; break;
				case STRETCH_INTEGER_XY :  stretch_request = STRETCH_INTEGER_X_FRACTIONAL_Y; break;
				case STRETCH_INTEGER_X_FRACTIONAL_Y : stretch_request = STRETCH_FRACTIONAL_XY; break;
				case STRETCH_FRACTIONAL_XY : stretch_request = STRETCH_NONE; break;
			}
			strcpy(resolution_request,"auto");
			change_state();
			/* tell updatescreen() to clean after us */
			schedule_full_refresh();
		}
	}

	if (input_ui_pressed(IPT_UI_CANCEL))
		sel = -1;

	if (input_ui_pressed(IPT_UI_CONFIGURE))
		sel = -2;

	if (sel == -1 || sel == -2)
	{
		/* tell updatescreen() to clean after us */
		schedule_full_refresh();
	}

	return sel + 1;
}

